from __future__ import print_function
from __future__ import absolute_import

# ###################################################################
# Import packages
# ###################################################################

import struct # for binary representation
import socket # for hostname identification
import string # for ascii selection
from datetime import datetime, timedelta, timezone
from twisted.protocols.basic import LineReceiver
from twisted.python import log
from martas.core import methods as mm


## Lippmann protocol
##
class LmProtocol(LineReceiver):
    """
    Protocol to read Lippmann HTRM Sensor data from serial unit
    Requires an init protocol

    Too be done

    """

    def __init__(self, client, sensordict, confdict):
        """
        'client' could be used to switch between different publishing protocols
                 (e.g. MQTT or WS-MCU gateway factory) to publish events
        'sensordict' contains a dictionary with all sensor relevant data (sensors.cfg)
        'confdict' contains a dictionary with general configuration parameters (martas.cfg)
        """
        self.client = client
        self.sensordict = sensordict    
        self.confdict = confdict
        self.count = 0  ## counter for sending header information
        self.sensor = sensordict.get('sensorid')
        self.hostname = socket.gethostname()
        self.printable = set(string.printable)
        #log.msg("  -> Sensor: {}".format(self.sensor))
        self.datalst = []
        self.datacnt = 0
        self.metacnt = 10
        # QOS
        self.qos=int(confdict.get('mqttqos',0))
        if not self.qos in [0,1,2]:
            self.qos = 0
        log.msg("  -> setting QOS:", self.qos)


    def connectionMade(self):
        log.msg('  -> {} connected.'.format(self.sensor))

    def connectionLost(self, reason):
        log.msg('  -> {} lost.'.format(self.sensor))


    def processData(self, data):
        """Convert raw ADC counts into SI units as per datasheets"""

        currenttime = datetime.now(timezone.utc).replace(tzinfo=None)
        # Correction for ms time to work with databank:
        currenttime_ms = currenttime.microsecond/1000000.
        ms_rounded = round(float(currenttime_ms),3)
        if not ms_rounded >= 1.0:
            currenttime = currenttime.replace(microsecond=int(ms_rounded*1000000.))
        else:
            currenttime = currenttime.replace(microsecond=0) + timedelta(seconds=1.0)
        filename = datetime.strftime(currenttime, "%Y-%m-%d")
        actualtime = datetime.strftime(currenttime, "%Y-%m-%dT%H:%M:%S.%f")
        lastActualtime = currenttime
        outtime = datetime.strftime(currenttime, "%H:%M:%S")
        timestamp = datetime.strftime(currenttime, "%Y-%m-%d %H:%M:%S.%f")

        sensorid = self.sensor
        packcode = '6hLL'
        header = "# MagPyBin %s %s %s %s %s %s %d" % (self.sensor, '[f]', '[f]', '[nT]', '[1000]', packcode, struct.calcsize('<'+packcode))

        try:
            intval = data[1].split(',')
            value = float(intval[0].strip())
            if 10000 < value < 100000:
                intensity = value
            else:
                intensity = 88888.0
        except ValueError:
            log.err("CS - Protocol: Not a number. Instead found:", data[0])
            intensity = 88888.0

        try:
            datearray = mm.time_to_array(timestamp)
            datearray.append(int(intensity*1000))
            data_bin = struct.pack('<'+packcode,*datearray)
        except:
            log.msg('Error while packing binary data')

        if not self.confdict.get('bufferdirectory','') == '':
            mm.data_to_file(self.confdict.get('bufferdirectory'), sensorid, filename, data_bin, header)

        return ','.join(list(map(str,datearray))), header


    def lineReceived(self, line):

        topic = self.confdict.get('station') + '/' + self.sensordict.get('sensorid')
        # extract only ascii characters
        line = ''.join(filter(lambda x: x in string.printable, str(line)))
        print ("line", line)
        """
        try:
            data = line.split()
            if len(data) == 2:
                dataarray, head = self.processData(data)
            elif len(data) == 1 and len(data.split(',')) == 2:
                data = ['$',data]
                dataarray, head = self.processData(data)
            else:
                log.msg('{}: Data seems not be appropriate data. Received data looks like: {}'.format(self.sensordict.get('protocol'),line))

            senddata = False
            coll = int(self.sensordict.get('stack'))
            if coll > 1:
                self.metacnt = 1 # send meta data with every block
                if self.datacnt < coll:
                    self.datalst.append(dataarray)
                    self.datacnt += 1
                else:
                    senddata = True
                    dataarray = ';'.join(self.datalst)
                    self.datalst = []
                    self.datacnt = 0
            else:
                senddata = True

            print ("Dataarray", dataarray, senddata, topic)
            if senddata:
                self.client.publish(topic+"/data", dataarray, qos=self.qos)
                if self.count == 0:
                    ## 'Add' is a string containing dict info like: 
                    ## SensorID:ENV05_2_0001,StationID:wic, PierID:xxx,SensorGroup:environment,... 
                    add = "SensorID:{},StationID:{},DataPier:{},SensorModule:{},SensorGroup:{},SensorDecription:{},DataTimeProtocol:{}".format( self.sensordict.get('sensorid',''),self.confdict.get('station',''),self.sensordict.get('pierid',''),self.sensordict.get('protocol',''),self.sensordict.get('sensorgroup',''),self.sensordict.get('sensordesc',''),self.sensordict.get('ptime','') )
                    self.client.publish(topic+"/dict", add, qos=self.qos)
                    self.client.publish(topic+"/meta", head, qos=self.qos)
                self.count += 1
                if self.count >= self.metacnt:
                    self.count = 0            
        except:
            log.err('{}: Unable to parse data {}'.format(self.sensordict.get('protocol'), line))
        """




# ###################################################################
# Import packages
# ###################################################################

import struct # for binary representation
import socket # for hostname identification
import string # for ascii selection
from datetime import datetime, timezone
from twisted.python import log
from martas.core import methods as mm
import serial
import sys

def send_command_ascii(ser,command,eol):
    #use printable here
    response = ''
    command = eol+command+eol
    if(ser.isOpen() == False):
        ser.open()
    sendtime = datetime.now(timezone.utc).replace(tzinfo=None)
    # encode to binary if python3
    if sys.version_info >= (3, 0):
        ser.write(command.encode('ascii'))
    else:
        ser.write(command)
    # skipping all empty lines
    while response == '':
        response = ser.readline()
    #responsetime = datetime.now(timezone.utc).replace(tzinfo=None)
    # decode from binary if py3
    if sys.version_info >= (3, 0):
        response = response.decode('ascii')
    responsetime = datetime.now(timezone.utc).replace(tzinfo=None)
    # return only ascii
    line = ''.join(filter(lambda x: x in string.printable, response))
    line = line.strip()
    return line, responsetime


def datetime2array(t):
    return [t.year,t.month,t.day,t.hour,t.minute,t.second,t.microsecond]

## Arduino active request protocol
## --------------------

class DSPProtocol(object):
    """
    Protocol to read DSP Winddata

    Active protocol

    Example sensor.cfg
    ULTRASONICDSP_0009009195_0001,S0,115200,8,1,N,active,None,60,1,DSP,ULTRASONICDSP,0009009195,0001,...

    """

    ## need a reference to our WS-MCU gateway factory to dispatch PubSub events
    ##
    def __init__(self, client, sensordict, confdict):
        # Specific commands for ultrasonic wind sensor
        # self.addressanemo = '01'
        self.commands = [{'data':'11TR00005'},{'meta':''},{'head':''}] # ,self.addressanemo+'TR00002']

        self.commands = [{'data':'01TR00002','c1':'01SH','c2':'01SL'}]
        self.hexcoding = False
        self.eol = '\r'


        self.client = client #self.wsMcuFactory = wsMcuFactory
        self.sensordict = sensordict
        self.confdict = confdict
        self.count = 0  ## counter for sending header information

        self.sensorname = sensordict.get('name')
        self.sensorid = sensordict.get('sensorid')
        self.baudrate=int(sensordict.get('baudrate'))
        self.port = confdict['serialport']+sensordict.get('port')
        self.parity=sensordict.get('parity')
        self.bytesize=sensordict.get('bytesize')
        self.stopbits=sensordict.get('stopbits')
        self.timeout=2 # should be rate depended

        self.hostname = socket.gethostname()
        self.printable = set(string.printable)
        self.datalst = []
        self.datacnt = 0
        rate = sensordict.get('rate','0')
        try:
            rate = int(rate)
        except:
            rate = 1
        if rate < 5:
            self.metacnt = 10 # send header at reduced rate
        else:
            self.metacnt = 1
        self.counter = {}

        # QOS
        self.qos=int(confdict.get('mqttqos',0))
        if not self.qos in [0,1,2]:
            self.qos = 0
        log.msg("  -> setting QOS:", self.qos)

        # switch on debug mode
        debugtest = confdict.get('debug')
        self.debug = False
        if debugtest == 'True':
            log.msg('DEBUG - {}: Debug mode activated.'.format(self.sensordict.get('protocol')))
            self.debug = True    # prints many test messages
        else:
            log.msg('  -> Debug mode = {}'.format(debugtest))

        # ULTRASONIC specific
        #self.sensorid = ''
        self.serialnum = ''
        self.serial1 = ''
        self.serial2 = ''


    def processData(self, sensorid, line, ntptime):
        """processing and storing data - requires sensorid and meta info
           data looks like (TR00002):
           01.6 290 +14.8 0E*4F
           windspeed, winddirection, virtualtemperature, status*pruefsumme
        """
        # currenttime = datetime.now(timezone.utc).replace(tzinfo=None)
        outdate = datetime.strftime(ntptime, "%Y-%m-%d")
        filename = outdate
        header = ''
        datearray = datetime2array(ntptime)
        packcode = '6hLlll'
        multiplier = [10,10,1]
        #print ("Processing line for {}: {}".format(sensorid, line))
        vals = line.split()
        if len(vals) > 3:
            errormark = False
            try:
                datearray.append(int(float(vals[2])*10))
                datearray.append(int(float(vals[0])*10))
                datearray.append(int(float(vals[1])*1))
            except:
                log.msg('{} protocol: Error while appending data to file - data looks like {}'.format(self.sensordict.get('protocol'),vals))
                errormark = True

            try:
                data_bin = struct.pack('<'+packcode,*datearray) #little endian
            except:
                log.msg('{} protocol: Error while packing binary data'.format(self.sensordict.get('protocol')))
                errormark = True

            if not errormark:
                #asseble header from available global information - write only if information is complete
                key = '[t2,var1,var2]'
                ele = '[Tv,V,Dir]'
                unit = '[degC,m_s,deg]'
                multplier = str(multiplier).replace(" ","")
                # Correct some common old problem
                unit = unit.replace('deg C', 'degC')
                #print ("ID process", sensorid)

                header = "# MagPyBin {} {} {} {} {} {} {}".format(sensorid, key, ele, unit, multplier, packcode, struct.calcsize('<'+packcode))
                data = ','.join(list(map(str,datearray)))

                if not self.confdict.get('bufferdirectory','') == '':
                    mm.data_to_file(self.confdict.get('bufferdirectory'), sensorid, filename, data_bin, header)
            else:
                data = ''
                header = ''
        else:
            data = ''

        return data, header


    def sendRequest(self):

        success = True
        # connect to serial
        ser = serial.Serial(self.port, baudrate=int(self.baudrate), parity=self.parity, bytesize=int(self.bytesize), stopbits=int(self.stopbits), timeout=int(self.timeout))
        for commdict in self.commands:
            for item in sorted(commdict):
                comm = commdict.get(item)
                if self.debug:
                    print ("sending item {} with command {}".format(item, comm))
                # eventually more frequently ask for 'data'
                answer, actime = send_command_ascii(ser,comm,self.eol)
                # disconnect from serial
                ser.close()
                if self.debug:
                    print ("got answer: {}".format(answer))
                self.serialnum = self.serial1+self.serial2
                if item == 'data': #  and len(self.serialnum) > 7: # removed serial number check as serial number is taken from sensors.cfg
                    #sensorid = "{}_{}_0001".format(str(self.sensorname),str(self.serialnum))
                    sensorid = self.sensorid
                    if self.debug:
                        print ("Processing data for {}". format(sensorid))
                    data, head = self.processData(sensorid, answer, actime)
                    # send data via mqtt
                    if not data == '':
                        self.sendmqtt(sensorid,data,head)
                if item == 'c2':
                    self.serial2 = answer.replace('!'+comm,'').strip()
                if item == 'c1':
                    self.serial1 = answer.replace('!'+comm,'').strip()

                # ##########################################################
                # answer is a string block with eventually multiple lines
                # apply the old linereceived method to each line of the answer
                # thats it ...

                #answer = ''.join(filter(lambda x: x in string.printable, answer))
                """
                if not success and self.errorcnt < 5:
                    self.errorcnt = self.errorcnt + 1
                    log.msg('SerialCall: Could not interpret response of system when sending %s' % item) 
                elif not success and self.errorcnt == 5:
                    try:
                        check_call(['/etc/init.d/martas', 'restart'])
                    except subprocess.CalledProcessError:
                        log.msg('SerialCall: check_call didnt work')
                        pass # handle errors in the called executable
                    except:
                        log.msg('SerialCall: check call problem')
                        pass # executable not found
                    #os.system("/etc/init.d/martas restart")
                    log.msg('SerialCall: Restarted martas process')
                """
        # disconnect from serial
        ser.close()

        # get answer
        #print("All commands send")

    def sendmqtt(self,sensorid,data,head):
        """
            call this method after processing data
        """

        ok = True
        if ok:

            topic = self.confdict.get('station') + '/' + sensorid

            senddata = False
            try:
                coll = int(self.sensordict.get('stack'))
            except:
                coll = 0

            if coll > 1:
                self.metacnt = 1 # send meta data with every block
                if self.datacnt < coll:
                    self.datalst.append(data)
                    self.datacnt += 1
                else:
                    senddata = True
                    data = ';'.join(self.datalst)
                    self.datalst = []
                    self.datacnt = 0
            else:
                senddata = True

            if senddata:
                if self.debug:
                    print ("publishing...")
                self.client.publish(topic+"/data", data, qos=self.qos)
                # If multiple sensors are connected, self.count needs to be a dict (with topic)
                # Initialize counter for each topic
                if self.counter.get(sensorid,'') == '':
                    self.counter[sensorid] = 0
                cnt = self.counter.get(sensorid)

                if cnt == 0:
                    ## 'Add' is a string containing dict info like: 
                    ## SensorID:ENV05_2_0001,StationID:wic, PierID:xxx,SensorGroup:environment,... 
                    add = "SensorID:{},StationID:{},DataPier:{},SensorModule:{},SensorGroup:{},SensorDescription:{},DataTimeProtocol:{}".format( sensorid,self.confdict.get('station','').strip(),self.sensordict.get('pierid','').strip(),self.sensordict.get('protocol','').strip(),self.sensordict.get('sensorgroup','').strip(),self.sensordict.get('sensordesc','').strip(),self.sensordict.get('ptime','').strip() )
                    self.client.publish(topic+"/dict", add, qos=self.qos)
                    self.client.publish(topic+"/meta", head, qos=self.qos)

                cnt += 1
                if cnt >= self.metacnt:
                    cnt = 0
                # update counter in dict 
                self.counter[sensorid] = cnt




#!/usr/bin/env python
# coding=utf-8

"""
MARTAS - threshold state machine
################################

DESCRIPTION:
Threshold state machine facilitates user defined finite state machines
to set defined actions in dependecy of data behavior.

Data is read from a defined source (DB or file, eventually MQTT).
Certain criteria e.g. when thresholds are exceeded or undergone will change
the state and create an action like sending an email or actuate a switch.
Following states can prevent further actions and wait for signals that
e.g. reset the state machine into the initial state.

statemachine.py can be scheduled in crontab.


REQUIREMENTS:
pip install geomagpy (>= 0.3.99)

APPLICATION:
statemachine -m /path/to/statemachine.cfg

OPTIONS:
    -l      List states of all state machines
    -r nr   Reset state machine number nr into initial state

statemachine.cfg: (looks like)
##  ----------------------------------------------------------------
##           CONFIGURATION DATA for STATEMACHINE.PY
##  ----------------------------------------------------------------

# run a statemachine best by a cron job
#   python path_to_MARTAS/app/statemachine.py -m this_config_file 

# MARTAS directory
martasdir            :   /home/cobs/MARTAS/

# Define data source (file, db, ...)
source               :   file

# If source = db then define data base credentials created by addcred (MARTAS)
dbcredentials        :   None

# If source = file define the MARTAS buffer base path
bufferpath           :   /srv/mqtt

# statusfile (a json style dictionary, which contains states) 
statusfile            :   /var/log/magpy/status.log

# Path of mail config file
emailconfig : /etc/martas/mail.cfg

# serial communication for switch commands (based on ardcomm.py (MARTAS/app))
serialcfg            :   None

# now the parameters of the state machine
# Example 2 shows how to avoid too much emails
#Nr. : 'sensorid';'timerange';'key';'value';'function';'operator';'statusmessage';'nextstatus'[;action;argument;action;argument...]
# comment: timerange in seconds, statusmessage is until now a dummy
status : start
1  :  SENSOR1_SERNR_0001;180;t1;20;min;below;;triggered;email;t1 below 20°
2  :  SENSOR1_SERNR_0001;180;t1;20;min;below;;triggered;email;t1 below 20°
status : triggered
1  :  SENSOR1_SERNR_0001;180;t1;21;min;above;;start;email;t1 above 21° - I like to send a lot of emails inspite of hystereses;-)
2  :  SENSOR1_SERNR_0001;180;t1;20;min;above;;reentered;email;t1 above 20° again - but that doesn't mean everything is fine again...
status : reentered
2  :  SENSOR1_SERNR_0001;120;t1;20;min;below;;stop;email;it seems like t1 is toggling around 20°C
status : stop
2  :  SENSOR1_SERNR_0001;120;t1;20;average;below;;stop;email;no more email.
"""


# Define packges to be used (local refers to test environment)
# ------------------------------------------------------------
from magpy.stream import DataStream, KEYLIST, NUMKEYLIST, read
from magpy.database import mysql,readDB
from datetime import datetime, timedelta
import magpy.opt.cred as mpcred
import sys, getopt, os
import json, copy

try:
    import paho.mqtt.client as mqtt
except:
    print ("MQTT not available")


if sys.version.startswith('2'):
    pyvers = '2'
else:
    pyvers = '3'


valuenamelist = ['sensorid','timerange','key','value','function','operator','statusmessage','nextstatus']


def is_number(s):
    try:
        float(s)
        return True
    except ValueError:
        return False
    except:
        return False

def readConfig(path):
    """
    DESCRIPTION:
        read configuration paths etc from a file
        this is stored in configdict
        read parameters and states of the state machines (1,2,3,...)
        complete info and behavior of the state machines is stored in statusdict
    """
    #try:
    if 1:
        status = ""
        # TODO backward compatibility? (-> default status, if not mentioned in the config file...)
        statusdict = {}
        configdict = {}
        config = open(path,'r')
        confs = config.readlines()

        for conf in confs:
            conflst = conf.split(' : ')
            if conf.startswith('#'): 
                continue
            elif conf.isspace():
                continue
            elif len(conflst) == 2:
                if conflst[0] == "status":
                    # get the status name for the following list
                    status = conflst[1].strip()
                if is_number(conflst[0]):
                    # extract parameterlist
                    key = conflst[0].strip()
                    values = conflst[1].strip().split(';')
                    valuedict = {}
                    if len(values) < len(valuenamelist):
                        # there must be entries in every field, additionally actions with args are possible
                        print ("PARAMETER: provided values differ from the expected amount - please check")
                    else:
                        for idx,valuename in enumerate(valuenamelist):
                            # mandatory parameters
                            valuedict[valuename] = values[idx].strip()
                        isAction = True
                        actionlist = []
                        for idx in range(len(valuenamelist),len(values)):
                            # optional parameters
                            if isAction:
                                action = values[idx].strip()
                                actionlist.append({'action':action})
                                isAction = False
                            else:
                                argument = values[idx].strip()
                                actionlist[-1]['argument'] = argument
                                isAction = True
                        if not actionlist == []:
                            valuedict['action'] = actionlist
                            # TODO important that len of action = argument?
                            #valuedict['argument'] = argumentlist
                        if not status in statusdict:
                            statusdict[status] = {}
                        if not key in statusdict[status]:
                            statusdict[status][key] = [copy.deepcopy(valuedict)]
                        else:
                            statusdict[status][key].append(copy.deepcopy(valuedict))
                else:
                    key = conflst[0].strip()
                    value = conflst[1].strip()
                    configdict[key] = value
    # status is not needed here
    del configdict['status']
    #except:
    if 0:
        print ("Problems when loading conf data from file...")
        #return ({}, {})
        print ("Could not obtain configuration data - aborting")
        sys.exit()

    return (configdict, statusdict)



def GetData(source, path, db, dbcredentials, sensorid, amount, startdate=None, debug=False):
    """
    DESCRIPTION:
    read the appropriate amount of data from the data file, database or mqtt stream
    """
 
    data = DataStream()
    msg = ''
    if not startdate:
        startdate = datetime.utcnow()
        endtime = None
    else:
        endtime = startdate

    starttime = startdate-timedelta(seconds=int(amount))

    if source in ['file','File']:
        filepath = os.path.join(path,sensorid)
        # TODO eventually check for existance and look for similar names
        # expath = CheckPath(filepath)
        # if expath:
        if debug:
            print ("Trying to access files in {}: Timerange: {} to {}".format(filepath,starttime,endtime))
        try:
            data = read(os.path.join(filepath,'*'), starttime=starttime, endtime=endtime)
        except:
            msg = "Could not access data for sensorid {}".format(sensorid)
            if debug:
                print (msg)
    elif source in ['db','DB','database','Database']:
        db = mysql.connect()
        data = readDB(db, sensorid, starttime=starttime)        

    if debug:
        print ("Got {} datapoints".format(data.length()[0]))

    return (data, msg) 

def GetTestValue(data=DataStream(), key='x', function='average', debug=False):
    """
    DESCRIPTION
    Returns comparison value(e.g. mean, max etc)
    """
    if debug:
        print ("Obtaining test value for key {} with function {}".format(key,function))
    func = 'mean'
    testvalue = None
    msg = ''
    n = data.length()[0]
    keys = data._get_key_headers()

    if not key in keys:
        print ("Requested key not found")
        return (testvalue, 'failure')
    if function in ['mean','Mean','average', 'Average','Median','median']:
        if n < 3:
            print ("not enough data points --- {} insignificant".format(function))
        if function in ['mean','Mean','average', 'Average']:
            func = 'mean'
        elif function in ['Median','median']:
            func = 'median'
        testvalue = data.mean(key,meanfunction=func)
    elif function in ['max','Max']:
        testvalue = data._get_max(key)
    elif function in ['min','Min']:
        testvalue = data._get_min(key)
    elif function in ['stddev','Stddev']:
        mean,testvalue = data.mean(key,std=True)
    else:
        msg = 'selected test function not available'

    if debug:
        print (" ... got {}".format(testvalue))

    return (testvalue, msg)


def CheckThreshold(testvalue, threshold, operator, debug=False):
    """
    DESCRIPTION:
     returns statusmessage
    """
    evaluate = False
    msg = ''
    if operator in ['below','Below','smaller']:
        comp = '<'
    elif operator in ['above','Above','greater']:
        comp = '>'
    elif operator in ['equal','Equal']:
        comp = '=='
    elif operator in ['equalabove']:
        comp = '>='
    elif operator in ['equalbelow']:
        comp = '<='
    else:
        msg = 'operator needs to be one of below, above or equal'
    tester = "{} {} {}".format(testvalue,comp,threshold)
    if debug:
        print ("Checking: {}".format(tester))
    try:
        if eval(tester):
            evaluate = True
            if debug:
                print (" ... positive")
        else:
            evaluate = False
            if debug:
                print (" ... negative - criteria not met")
    except:
        msg = "Comparison {} failed".format(tester)

    return (evaluate, msg)

def InterpreteStatus(valuedict, debug=False):
    """
    DESCRIPTION:
    checks the message and replace certain keywords with predefined test
    """
    msg = valuedict.get('statusmessage')
    defaultline =  'Current {} {} {}'.format(valuedict.get('function'),valuedict.get('state'),valuedict.get('value'))
    ct = datetime.utcnow()
    msg = msg.replace('none', '')
    msg = msg.replace('default', defaultline)
    msg = msg.replace('date', datetime.strftime(ct,"%Y-%m-%d"))
    msg = msg.replace('hour', datetime.strftime(ct,"%Y-%m-%d %H:%M"))
    msg = msg.replace('minute', datetime.strftime(ct,"%Y-%m-%d %H:%M"))
    msg = msg.replace('month', datetime.strftime(ct,"%Y-%m"))
    msg = msg.replace('year', datetime.strftime(ct,"%Y"))
    msg = msg.replace('week', 'week {}'.format(ct.isocalendar()[1]))

    return msg


def readConfigFromFile(path):
    """
    read out a file into a dict
    as generally as possible

    how a config file should look like:
    ##  -------------------------------
    # # are comments

    key  :  value

    # there must be at least one ' ' (space) beside the ':'
    # so the divider is ' : '
    ##  -------------------------------
    """

    dic = {}
    #try:
    if 1:
        config = open(path,'r')
        confs = config.readlines()

        for conf in confs:
            conflst = conf.split(' : ')
            if conf.startswith('#'):
                continue
            elif conf.isspace():
                continue
            elif len(conflst) == 2:
                key = conflst[0].strip()
                value = conflst[1].strip()
                dic[key] = value
    #except:
    #    print ("Problems when loading conf data from file.")

    return dic



def main(argv):

    #para = sp.parameterdict
    #conf = sp.configdict
    para = {}
    conf = {}
    # necessary configs (may be overwritten):
    conf['statusfile']='/var/log/magpy/statusfile.log'
    debug = False
    configfile = None
    statusdict = {}
    statuskeylist = []
    MachineToReset = None
    ListMachine = False

    usagestring = 'threshold.py -h <help> -m <configpath> [-l][-r state machine number]'
    try:
        opts, args = getopt.getopt(argv,"hm:Ur:l",["configpath=","reset="])
    except getopt.GetoptError:
        print ('Check your options:')
        print (usagestring)
        sys.exit(2)
    for opt, arg in opts:
        if opt == '-h':
            print ('------------------------------------------------------')
            print ('Usage:')
            print (usagestring)
            print ('------------------------------------------------------')
            print ('Options:')
            print ('-h            help')
            print ('-l            List states of all state machines')
            print ('-r nr         Reset state machine number nr into initial state')
            print ('-m            Define the path for the configuration file.')
            print ('              Please note: a configuration file is obligatory')
            print ('              ----------------------------')
            print ('              configurationfile')
            print ('              ----------------------------')
            print ('              statemachine.cfg: (looks like)')
            print ('              # MARTAS directory')
            print ('              martasdir            :   /home/cobs/MARTAS/')
            print ('              # Define data source (file, db, ...)')
            print ('              source               :   file')
            print ('              # If source = db then define data base credentials created by addcred (MARTAS)')
            print ('              dbcredentials        :   None')
            print ('              # If source = file define the MARTAS buffer base path')
            print ('              bufferpath           :   /srv/mqtt/')
            print ('              # statusfile (a json style dictionary, which contains states)')
            print ('              statusfile           :   /var/log/magpy/status.log')
            print ('              # Path of mail config file')
            print ('              emailconfig : /etc/martas/mail.cfg')
            print ('              # serial communication for switch commands (based on ardcomm.py (MARTAS/app))')
            print ('              serialcfg            :   None')
            print ("              #Nr. : 'sensorid';'timerange';'key';'value';'function';'operator';'statusmessage';'nextstatus'[;action;argument;action;argument...]")
            print ('              # comment: timerange in seconds, statusmessage is until now a dummy')
            print ('              status : start')
            print ('              1  :  SENSOR1_SERNR_0001;180;t1;20;min;below;;triggered;email;t1 below 20°')
            print ('              2  :  SENSOR1_SERNR_0001;180;t1;20;min;below;;triggered;email;t1 below 20°')
            print ('              status : triggered')
            print ('              1  :  SENSOR1_SERNR_0001;180;t1;21;min;above;;start;email;t1 above 21° - I like to send a lot of emails inspite of hystereses;-)')
            print ("              2  :  SENSOR1_SERNR_0001;180;t1;20;min;above;;reentered;email;t1 above 20° again - but that doesn't mean everything is fine again...")
            print ('              status : reentered')
            print ('              2  :  SENSOR1_SERNR_0001;120;t1;20;min;below;;stop;email;it seems like t1 is toggling around 20°C')
            print ('              2  :  SENSOR1_SERNR_0001;120;t1;20;average;below;;stop;email;no more email')
            print ('              ----------------------------')
            print ('')
            print ('-r            reset a state machine')
            print ('-l            display states')
            print ('------------------------------------------------------')
            print ('Example:')
            print ('   python statemachine.py -m /etc/martas/statemachine.cfg')
            sys.exit()
        elif opt in ("-m", "--configfile"):
            configfile = arg
            if debug:
                print ("Getting all parameters of the state machine from configuration file: {}".format(configfile))
            (conf, para) = readConfig(configfile)
        elif opt in ("-r", "--reset"):
            if is_number(arg):
                MachineToReset = arg
            else:
                print ("--reset must_be_a_number")
                sys.exit()
        elif opt in ("-l"):
            ListMachine = True
        elif opt in ("-U", "--debug"):
            debug = True

    if debug:
        print ("Configuration dictionary: \n{}".format(conf))
        print ("Parameter dictionary: \n{}".format(para))

    if not (len(para)) > 0:
        print ("No parameters given to be checked - aborting") 
        sys.exit()


    try:
        martaslogpath = os.path.join(conf.get('martasdir'), 'core')
        sys.path.insert(1, martaslogpath)
        import martas
        logpath = conf.get('bufferpath')
    except:
        print ("Could not import martas logging routines - check MARTAS directory path")


    statusdict = {}
    if os.path.isfile(conf['statusfile']):
        # read log if exists and exentually update changed information
        # return changes
        with open(conf['statusfile'], 'r') as file:
            statusdict = json.load(file)
        if debug:
            print ("Statusfile {} loaded".format(conf['statusfile']))

    if ListMachine:
        if statusdict == {}:
            print ('no states but start')
        else:
            for state in statusdict:
                print (state+": "+statusdict[state]['status'])
        exit()

    if MachineToReset and not statusdict == {}:
        if MachineToReset in statusdict:
            del statusdict[MachineToReset]
            print (MachineToReset+" set to 'start'")
        else:
            print (MachineToReset+" not found")
        with open(conf['statusfile'], 'w') as file:
            if debug:
                print ('writing to '+conf['statusfile']+' :')
                print (statusdict)
            file.write(json.dumps(statusdict)) # use `json.loads` to do the reverse
        exit()
    
    # For each machine
    for i in range(0,1000): 
        valuedict = {}
        values = []
        if not str(i) in statusdict and str(i) in para['start']:
            # machine is not yet in the statusfile, let's add it to the dict
            statusdict[str(i)] = {}
            statusdict[str(i)]['status'] = 'start'
            #laststatusdict[str(i)]['sensorid'] = para['start'][str(i)]['sensorid']
            #laststatusdict[str(i)]['key'] = para['start'][str(i)]['key']
        if str(i) in statusdict:
            status = statusdict[str(i)]['status']
            # TODO handle states deleted from the config file!
            values = para[status].get(str(i),[])
        if not values == []:
            if debug:
                print ("Checking state machine {}".format(i))
            data = DataStream()
                
            # Obtain a magpy data stream of the respective data set

            for valuedict in values:
                if debug:
                    print ("Accessing data from {} at {}: Sensor {} - Amount: {} sec".format(conf.get('source'),conf.get('bufferpath'),valuedict.get('sensorid'),valuedict.get('timerange') ))
                (data,msg1) = GetData(conf.get('source'), conf.get('bufferpath'), conf.get('database'), conf.get('dbcredentials'), valuedict.get('sensorid'),valuedict.get('timerange'), debug=debug , startdate=conf.get('startdate') )
                testvalue = None
                if data._get_key_headers() == []:
                    # there are no keys in the data
                    if debug:
                        print ('no data for testvalue')
                else:
                    (testvalue,msg2) = GetTestValue( data, valuedict.get('key'), valuedict.get('function'), debug=debug) # Returns comparison value(e.g. mean, max etc)
                if debug:
                    print ("testvalue is {}".format(testvalue))
                if is_number(testvalue):
                    (evaluate, msg) = CheckThreshold(testvalue, valuedict.get('value'), valuedict.get('operator'), debug=debug) # Returns statusmessage
                    if evaluate and msg == '':
                        # criteria are met - do something
                        # change status
                        if debug:
                            print ("changing status of machine "+str(i)+" from")
                            print (statusdict[str(i)]['status'])
                            print ("to")
                            print (valuedict['nextstatus'])
                        statusdict[str(i)]['status'] = valuedict['nextstatus']
                        if 'action' in valuedict:
                            for action in valuedict['action']:
                                if action['action'] == 'email':
                                    dic = readConfigFromFile(conf.get('emailconfig'))
                                    dic['Text'] = action['argument']
                                    martas.sendmail(dic)
                                # TODO not implemented / not tested
                                if action['action'] == 'telegram':
                                    dic = sm_support.readConfigFromFile(conf.get('telegramconfig'))
                                    dic['text'] = action['argument']
                                    martas.sendtelegram(dic)
                                if action['action'] == 'switch:':
                                    dic = conf
                                    dic['comm'] = action['argument']
                                    martas.sendswitchcommand(dic)



                        # TODO handle content resp. errors
                        content = InterpreteStatus(valuedict,debug=debug)
                        # Perform switch and added "switch on/off" to content 
                        if not valuedict.get('switchcommand') in ['None','none',None]:
                            if debug:
                                print ("Found switching command ... eventually will send serial command (if not done already) after checking all other commands")
                            content = '{} - switch: {}'.format(content, valuedict.get('switchcommand'))
                            # remember the switchuing command and only issue it if statusdict is changing
                    elif not msg == '':
                        content =  msg
                    else:
                        content = ''
            else:
                #content = msg1+' - '+msg2
                pass

            #if content:
            if 0:
                statuskeylist.append('Sensor {} and key {}'.format(valuedict.get('sensorid'),valuedict.get('key')))
                statusdict['Sensor {} and key {}'.format(valuedict.get('sensorid'),valuedict.get('key'))] = content

            if debug:
                print ("Finished state machine {}".format(i))

    with open(conf['statusfile'], 'w') as file:
        if debug:
            print ('writing to '+conf['statusfile']+' :')
            print (statusdict)
        file.write(json.dumps(statusdict)) # use `json.loads` to do the reverse
    exit()
    

    if conf.get('reportlevel') == 'full':
        # Get a unique status key list:
        statuskeylist = list(dict.fromkeys(statuskeylist))
        for elem in statuskeylist:
            cont = statusdict.get(elem,'')
            if cont == '':
                statusdict[elem] = "Everything fine"

    if debug:
        print ("Statusdict: {}".format(statusdict))

    receiver = conf.get('notification')
    cfg = conf.get('notificationconfig')
    logfile = conf.get('logfile')

    if debug:
        print ("New notifications will be send to: {} (Config: {})".format(receiver,cfg))

    martaslog = ml(logfile=logfile,receiver=receiver)
    if receiver == 'telegram':
        martaslog.telegram['config'] = cfg
    elif receiver == 'email':
        martaslog.email['config'] = cfg

    changes = martaslog.msg(statusdict)

    if not len(changes) > 0:
        print ("Nothing to report - threshold check successfully finished")

    for element in changes:
        line = changes.get(element)
        if debug:
            print ("Changes affecting:", element)
        l = line.split('switch:')
        if len(l) == 2:
            print (" ... now dealing with switching serial command:")
            comm = l[1].strip()
            script = os.path.join(conf.get('martasdir'),'app','ardcomm.py')
            pythonpath = sys.executable
            arg1 = "-c {}".format(comm)
            arg2 = "-p {}".format(conf.get('port'))
            arg3 = "-b {}".format(conf.get('baudrate'))
            arg4 = "-a {}".format(conf.get('parity'))
            arg5 = "-y {}".format(conf.get('bytesize'))
            arg6 = "-s {}".format(conf.get('stopbits'))
            arg7 = "-t {}".format(conf.get('timeout'))
            #arg8 = "-e {}".format(conf.get('eol')) # not used so far

            command = "{} {} {} {} {} {} {} {} {}".format(pythonpath,script,arg1, arg2, arg3, arg4, arg5, arg6, arg7) ## Too be checked
            command = "{} {} {}".format(pythonpath,script,arg1)
            if debug:
                print (" ... sending {}".format(command))


            try:
                import subprocess
                p = subprocess.Popen(command, stdout=subprocess.PIPE, shell=True)
                (output, err) = p.communicate()
                mesg = "{}".format(output)
            except subprocess.CalledProcessError:
                mesg = "threshold: sending command didnt work"
            except:
                mesg = "threshold: sending command problem"

            print (mesg)

            print (" ... success")

if __name__ == "__main__":
   main(sys.argv[1:])





# ###################################################################
# Import packages
# ###################################################################

import struct # for binary representation
import socket # for hostname identification
from datetime import datetime, timezone, timedelta
from martas.core import methods as mm
from magpy.opt import cred as mpcred
from twisted.python import log

import dateutil.parser as dparser
import urllib.request
import json 


class GICProtocol():
    """
    Protocol to read GIC URL data , data looks like
    [{"client":"gic01","date":"2022-04-13","timeUTC":"16:43:05","temperatureDegC":"34.8","NpcMilliAmps":"120.0434"},{"client":"gic02",... }]

    The protocol defines the sensor name in its init section, which
    is used to dispatch url links and define local storage folders

    URLs are then taken from magpy credentials by extracting the address for the given sensor name. So you first need
    to add access credentials as follows:
    addcred -t transfer -c gicaut -u USERNAME -p PASSWORD -a https://URL.TO.GIC.DATA:PORT

    Add the following line to sensors.cfg to request new data every 60 seconds:
    gicaut,URL,-,-,-,-,active,None,60,1,GIC,GIC,-,0001,-,different,NTP,spaceweather,geomagnetically induced currents


    """
    def __init__(self, client, sensordict, confdict):
        """
        'client' could be used to switch between different publishing protocols
                 (e.g. MQTT or WS-MCU gateway factory) to publish events
        'sensordict' contains a dictionary with all sensor relevant data (sensors.cfg)
        'confdict' contains a dictionary with general configuration parameters (martas.cfg)
        """
        self.client = client
        self.sensordict = sensordict
        self.confdict = confdict
        self.count = 0  ## counter for sending header information
        self.sensor = sensordict.get('sensorid')
        self.revision = sensordict.get('revision')
        self.hostname = socket.gethostname()
        #print ("  -> Sensor: {}".format(self.sensor))
        self.datalst = []
        self.datacnt = 0
        self.metacnt = 10
        self.url = mpcred.lc(self.sensor,'address')
        self.qos=int(confdict.get('mqttqos',0))
        if not self.qos in [0,1,2]:
            self.qos = 0
        log.msg("  -> setting QOS:", self.qos)
        self.debug = False

    def sendRequest(self):

        with urllib.request.urlopen(self.url) as url:
            data = json.loads(url.read().decode())
        self.dataReceived(data)
        #expect:
        #log.msg('  -> {} unavailable.'.format(self.sensor))


    def processData(self, data):
        """Process GIC URL data """

        if not data:
            log.msg('  -> {} - received empty data structure'.format(self.sensor))
            return {}
        currenttime = datetime.now(timezone.utc).replace(tzinfo=None)
        datadict = {}
        for el in data:
            #if self.debug:
            #print ("ANALYSIS", el)
            dataname = el.get("client")
            gicval = str(el.get("NpcMilliAmps")) # can be a single value, string with ; and an integer
            #print (gicval)
            giclist = gicval.split(";")
            sdate = el.get("date")
            stime = el.get("timeUTC")
            temperature = el.get("temperatureDegC")
            cnt = len(giclist)
            lines = []
            packcode = '6hLll'
            sensorid = "{}_{}_{}".format(self.sensor.upper(), dataname.upper(), self.revision)
            header = "# MagPyBin %s %s %s %s %s %s %d" % (sensorid, '[x,t2]', '[GIC,T]', '[mA,degC]', '[1000,1000]',
                                                          packcode, struct.calcsize('<' + packcode))
            for i,gic in enumerate(giclist):
                # ignore empty and experimental data sets
                gic = float(gic)
                if gic and not gic in [555000] and not dataname in ['gic20'] and not sdate in [555000]:
                    datearray = []
                    data_bin = None
                    if self.debug:
                        print (" i, len and datecorr", i, len(giclist), len(giclist)-(i+1))
                    datatime = dparser.parse("{} {}".format(sdate,stime))-timedelta(seconds=(len(giclist)-(i+1)))
                    ###
                    filename = datetime.strftime(currenttime, "%Y-%m-%d") # use PC time for buffername
                    timestamp = datetime.strftime(datatime, "%Y-%m-%d %H:%M:%S.%f")
                    ###

                    try:
                        gic = float(gic)
                    except:
                        gic = 999999
                    if gic > 554999 and gic < 555001:
                        gic = 999999
                    try:
                        temperature = float(temperature)
                    except:
                        temperature = 999999
                    try:
                        if self.debug:
                            print ("Writing:", sensorid, timestamp, gic, temperature)
                        if not gic==999999:
                            datearray = mm.time_to_array(timestamp)
                            datearray.append(int(gic*1000))
                            datearray.append(int(temperature*1000))
                            data_bin = struct.pack('<'+packcode,*datearray)  #use little endian byte order
                    except:
                        log.msg('Error while packing binary data')
                        pass

                    if not self.confdict.get('bufferdirectory','') == '' and data_bin:
                        mm.data_to_file(self.confdict.get('bufferdirectory'), sensorid, filename, data_bin, header)
                    if datearray:
                        lines.append(','.join(list(map(str,datearray))))
                #print (lines, header)
                datadict[sensorid] = {'lines': lines, 'head':header}

        return datadict

    def dataReceived(self, data):
        # extract data array and buffer it
        datadict = self.processData(data)
        for dataid in datadict:
            topic = self.confdict.get('station') + '/' + dataid
            datavals = datadict.get(dataid)
            datalines = datavals.get('lines')
            datahead = datavals.get('head')

            #ok=True
            #if ok:
            try:
                senddata = True

                if senddata:
                    if self.count >= self.metacnt:
                        self.count = 0
                    for dataline in datalines:
                        self.client.publish(topic+"/data", dataline, qos=self.qos)
                    if self.count == 0:
                        ## 'Add' is a string containing dict info like:
                        ## SensorID:ENV05_2_0001,StationID:wic, PierID:xxx,SensorGroup:environment,... 
                        add = "SensorID:{},StationID:{},DataPier:{},SensorModule:{},SensorGroup:{},SensorDescription:{},DataTimeProtocol:{}".format( self.sensordict.get('sensorid',''),self.confdict.get('station',''),self.sensordict.get('pierid',''),self.sensordict.get('protocol',''),self.sensordict.get('sensorgroup',''),self.sensordict.get('sensordesc',''),self.sensordict.get('ptime','') )
                        self.client.publish(topic+"/dict", add, qos=self.qos)
                        self.client.publish(topic+"/meta", datahead, qos=self.qos)
            except:
                log.err('{}: Unable to parse data {}'.format(self.sensordict.get('protocol'), data))

        self.count += 1




# ###################################################################
# Import packages
# ###################################################################

import re     # for interpretation of lines
import struct # for binary representation
import socket # for hostname identification
import string # for ascii selection
import sys
from datetime import datetime, timezone
from twisted.protocols.basic import LineReceiver
from twisted.python import log
from martas.core import methods as mm


class EnvProtocol(LineReceiver):
    """
    Protocol to read MessPC EnvironmentalSensor 5 data from usb unit
    Each sensor has its own class (that can be improved...)
    The protocol defines the sensor name in its init section, which
    is used to dipatch url links and define local storage folders

    """
    def __init__(self, client, sensordict, confdict):
        """
        'client' could be used to switch between different publishing protocols
                 (e.g. MQTT or WS-MCU gateway factory) to publish events
        'sensordict' contains a dictionary with all sensor relevant data (sensors.cfg)
        'confdict' contains a dictionary with general configuration parameters (martas.cfg)
        """
        self.client = client
        self.sensordict = sensordict
        self.confdict = confdict
        self.count = 0  ## counter for sending header information
        self.sensor = sensordict.get('sensorid')
        self.hostname = socket.gethostname()
        self.printable = set(string.printable)
        #log.msg("  -> Sensor: {}".format(self.sensor))
        self.datalst = []
        self.datacnt = 0
        self.metacnt = 10
        self.qos=int(confdict.get('mqttqos',0))
        if not self.qos in [0,1,2]:
            self.qos = 0
        log.msg("  -> setting QOS:", self.qos)
        # PYTHON version
        self.pvers = sys.version_info[0]

    def connectionMade(self):
        log.msg('  -> {} connected.'.format(self.sensor))

    def connectionLost(self, reason):
        log.msg('  -> {} lost.'.format(self.sensor))

    def processData(self, data):
        """Process Environment data """

        currenttime = datetime.now(timezone.utc).replace(tzinfo=None)
        outdate = datetime.strftime(currenttime, "%Y-%m-%d")
        filename = outdate
        actualtime = datetime.strftime(currenttime, "%Y-%m-%dT%H:%M:%S.%f")
        outtime = datetime.strftime(currenttime, "%H:%M:%S")
        timestamp = datetime.strftime(currenttime, "%Y-%m-%d %H:%M:%S.%f")
        packcode = '6hLllL'
        sensorid = self.sensor
        header = "# MagPyBin %s %s %s %s %s %s %d" % (sensorid, '[t1,t2,var1]', '[T,DewPoint,RH]', '[degC,degC,per]', '[1000,1000,1000]', packcode, struct.calcsize('<'+packcode))

        valrh = re.findall(r'\d+',data[0])
        if len(valrh) > 1:
            temp = float(valrh[0] + '.' + valrh[1])
        else:
            temp = float(valrh[0])
        valrh = re.findall(r'\d+',data[1])
        if len(valrh) > 1:
            rh = float(valrh[0] + '.' + valrh[1])
        else:
            rh = float(valrh[0])
        valrh = re.findall(r'\d+',data[2])
        if len(valrh) > 1:
            dew = float(valrh[0] + '.' + valrh[1])
        else:
            dew = float(valrh[0])

        try:
            datearray = mm.time_to_array(timestamp)
            datearray.append(int(temp*1000))
            datearray.append(int(dew*1000))
            datearray.append(int(rh*1000))
            data_bin = struct.pack('<'+packcode,*datearray)  #use little endian byte order
        except:
            log.msg('Error while packing binary data')
            pass

        if not self.confdict.get('bufferdirectory','') == '':
            mm.data_to_file(self.confdict.get('bufferdirectory'), sensorid, filename, data_bin, header)
        return ','.join(list(map(str,datearray))), header

    def lineReceived(self, line):
        topic = self.confdict.get('station') + '/' + self.sensordict.get('sensorid')
        if self.pvers > 2:
            line=line.decode('latin')
        line = ''.join(filter(lambda x: x in string.printable, str(line)))

        try:
            data = line.split()
            if len(data) == 3:
                data, head = self.processData(data)
            else:
                log.msg('{}: Data seems not be appropriate data. Received data looks like: {}'.format(self.sensordict.get('protocol'),line))

            senddata = False
            coll = int(self.sensordict.get('stack'))
            if coll > 1:
                self.metacnt = 1 # send meta data with every block
                if self.datacnt < coll:
                    self.datalst.append(data)
                    self.datacnt += 1
                else:
                    senddata = True
                    data = ';'.join(self.datalst)
                    self.datalst = []
                    self.datacnt = 0
            else:
                senddata = True

            if senddata:
                self.client.publish(topic+"/data", data, qos=self.qos)
                if self.count == 0:
                    ## 'Add' is a string containing dict info like:
                    ## SensorID:ENV05_2_0001,StationID:wic, PierID:xxx,SensorGroup:environment,...
                    add = "SensorID:{},StationID:{},DataPier:{},SensorModule:{},SensorGroup:{},SensorDescription:{},DataTimeProtocol:{}".format( self.sensordict.get('sensorid',''),self.confdict.get('station',''),self.sensordict.get('pierid',''),self.sensordict.get('protocol',''),self.sensordict.get('sensorgroup',''),self.sensordict.get('sensordesc',''),self.sensordict.get('ptime','') )
                    self.client.publish(topic+"/dict", add, qos=self.qos)
                    self.client.publish(topic+"/meta", head, qos=self.qos)
                self.count += 1
                if self.count >= self.metacnt:
                    self.count = 0
        except:
            log.err('{}: Unable to parse data {}'.format(self.sensordict.get('protocol'), line))



#!/usr/bin/env python
import dash
import dash_bootstrap_components as dbc


app = dash.Dash(
    __name__,
    use_pages=True,
    external_stylesheets=[dbc.themes.BOOTSTRAP],
)

navbar = dbc.NavbarSimple(
    dbc.Nav(
        [
            dbc.NavLink(page["name"], href=page["path"])
            for page in dash.page_registry.values()
            if page.get("top_nav")
        ],
    ),
    brand="MARTAS viewer",
    color="#888888",
    dark=True,
    className="mb-2",
)


app.layout = dbc.Container(
    [navbar, dash.page_container],
    fluid=True,
)


if __name__ == "__main__":
    app.run(host="0.0.0.0", debug=False)



# ###################################################################
# Import packages
# ###################################################################

import struct # for binary representation
import socket # for hostname identification
import string # for ascii selection
import sys
from datetime import datetime, timezone
from twisted.protocols.basic import LineReceiver
from twisted.python import log
from martas.core import methods as mm


## Caesium protocol
##
class CsProtocol(LineReceiver):
    """
    Protocol to read Geometrics CS Sensor data from serial unit
    Each sensor has its own class (that can be improved...)
    The protocol defines the sensor name in its init section, which
    is used to dipatch url links and define local storage folders

    """

    def __init__(self, client, sensordict, confdict):
        """
        'client' could be used to switch between different publishing protocols
                 (e.g. MQTT or WS-MCU gateway factory) to publish events
        'sensordict' contains a dictionary with all sensor relevant data (sensors.cfg)
        'confdict' contains a dictionary with general configuration parameters (martas.cfg)
        """
        self.client = client
        self.sensordict = sensordict
        self.confdict = confdict
        self.count = 0  ## counter for sending header information
        self.sensor = sensordict.get('sensorid')
        self.hostname = socket.gethostname()
        self.printable = set(string.printable)
        #log.msg("  -> Sensor: {}".format(self.sensor))
        self.datalst = []
        self.datacnt = 0
        self.metacnt = 10
        # QOS
        self.qos=int(confdict.get('mqttqos',0))
        if not self.qos in [0,1,2]:
            self.qos = 0
        log.msg("  -> setting QOS:", self.qos)
        # PYTHON version
        self.pvers = sys.version_info[0]


    def connectionMade(self):
        log.msg('  -> {} connected.'.format(self.sensor))

    def connectionLost(self, reason):
        log.msg('  -> {} lost.'.format(self.sensor))


    def processData(self, data):
        """Convert raw ADC counts into SI units as per datasheets"""

        datearray = []
        data_bin = None
        currenttime = datetime.now(timezone.utc).replace(tzinfo=None)
        filename = datetime.strftime(currenttime, "%Y-%m-%d")
        #actualtime = datetime.strftime(currenttime, "%Y-%m-%dT%H:%M:%S.%f")
        #lastActualtime = currenttime
        #outtime = datetime.strftime(currenttime, "%H:%M:%S")
        timestamp = datetime.strftime(currenttime, "%Y-%m-%d %H:%M:%S.%f")

        sensorid = self.sensor
        packcode = '6hLL'
        header = "# MagPyBin %s %s %s %s %s %s %d" % (self.sensor, '[f]', '[f]', '[nT]', '[1000]', packcode, struct.calcsize('<'+packcode))

        try:
            intval = data[1].split(',')
            value = float(intval[0].strip())
            if 10000 < value < 100000:
                intensity = value
            else:
                intensity = 88888.0
        except ValueError:
            log.err("CS - Protocol: Not a number. Instead found:", data[0])
            intensity = 88888.0

        try:
            datearray = mm.time_to_array(timestamp)
            datearray.append(int(intensity*1000))
            data_bin = struct.pack('<'+packcode,*datearray)
        except:
            log.msg('Error while packing binary data')

        if not self.confdict.get('bufferdirectory','') == '':
            mm.data_to_file(self.confdict.get('bufferdirectory'), sensorid, filename, data_bin, header)

        return ','.join(list(map(str,datearray))), header


    def lineReceived(self, line):

        topic = self.confdict.get('station') + '/' + self.sensordict.get('sensorid')
        line=line.decode('ascii', 'ignore')
        # extract only ascii characters
        line = ''.join(filter(lambda x: x in string.printable, str(line)))

        dataarray = None
        head = None
        ok = True
        try:
            data = line.split()
            if len(data) == 2:
                dataarray, head = self.processData(data)
            elif len(data) == 1 and len(data[0].split(',')) == 2:
                data = ['$',data]
                dataarray, head = self.processData(data)
            else:
                log.msg('{}: did not find appropriate data. Received data looks like: {}'.format(self.sensordict.get('protocol'),line))
                ok = False
        except:
            log.err('{}: Unable to parse data {}'.format(self.sensordict.get('protocol'), line))
            ok = False

        if ok and dataarray:
            senddata = False
            coll = int(self.sensordict.get('stack'))
            if coll > 1:
                self.metacnt = 1 # send meta data with every block
                if self.datacnt < coll:
                    self.datalst.append(dataarray)
                    self.datacnt += 1
                else:
                    senddata = True
                    dataarray = ';'.join(self.datalst)
                    self.datalst = []
                    self.datacnt = 0
            else:
                senddata = True

            #print ("Dataarray", dataarray, senddata, topic)
            if senddata:
                self.client.publish(topic+"/data", dataarray, qos=self.qos)
                if self.count == 0:
                    ## 'Add' is a string containing dict info like:
                    ## SensorID:ENV05_2_0001,StationID:wic, PierID:xxx,SensorGroup:environment,...
                    add = "SensorID:{},StationID:{},DataPier:{},SensorModule:{},SensorGroup:{},SensorDecription:{},DataTimeProtocol:{}".format( self.sensordict.get('sensorid',''),self.confdict.get('station',''),self.sensordict.get('pierid',''),self.sensordict.get('protocol',''),self.sensordict.get('sensorgroup',''),self.sensordict.get('sensordesc',''),self.sensordict.get('ptime','') )
                    self.client.publish(topic+"/dict", add, qos=self.qos)
                    self.client.publish(topic+"/meta", head, qos=self.qos)
                self.count += 1
                if self.count >= self.metacnt:
                    self.count = 0



#!/usr/bin/env python
"""
Regularly truncate contents of timesseries in a MagPy database
whereas "archive" also allows for truncating the database (based on DATAINO)
"db_truncate" removes contents from all tables of xxx_xxx_xxxx_xxxx structures
(independent of DATAINFO contents)
"""
from magpy.stream import *
from magpy.core import database
from magpy.opt import cred as mpcred

from martas.core.methods import martaslog as ml
from martas.core import methods as mm

from datetime import datetime, timezone
import getopt
import pwd
import socket


"""
#configuration:

credentials     :      cobsdb

# cleanratio = samplingrateratio for deleting old db entries - default is 12')
#                   deleting data older than samplingrate(sec)*12 days.')
#                    cleanratio=12 : 1sec data older than 12 days is deleted in DB')
#                                  :           1min data older than 720 days is deleted in DB')
#                    cleanratio=1  : 1sec data older than 1 day is deleted in DB')
#                                  :           1min data older than 60 days is deleted in DB
cleanratio      :      12

# Sensors present in path to be skipped (Begging of Sensorname is enough
blacklist       :    BLV,QUAKES,Sensor2,Sensor3,

# If sensorlist is provided only these sensors are considered
#sensorlist      :    Sensor1,Sensor4,



DESCRIPTION:
    db_truncate.py truncates contents of timesseries in a MagPy database.
    Whereas "archive" also allows for truncating the database (based on DATAINO)
    "db_truncate" removes contents from all tables of xxx_xxx_xxxx_xxxx structure.
    (independent of DATAINFO contents).
    The databank size is automatically restricted
    in dependency of the sampling rate of the input data. A cleanratio of 12
    will only keep the last
    12 days of second data, the last 720 days of minute data and
    approximately 118 years of hourly data are kept.
    Settings are given in a configuration file.

APPLICATION:
    # Automatic
    python3 db_truncate.py -c truncate.cfg

    # Manual for specific sensors and time range
    python3 db_truncate.py -c config.cfg -s Sensor1,Sensor2 -i 12

"""

def query_db(db, sql, debug=False):
    if not db:
        return []
    if debug:
         print ("   Sending sql query: {}".format(sql))
    cursor = db.db.cursor()
    message = db._executesql(cursor, sql)
    if message:
        print(message)
        return []
    return cursor.fetchall()

def get_table_tist(db, sensorlist=None, blacklist=None, debug=False):

    if not sensorlist:
        sensorlist=[]
    if not blacklist:
        blacklist=[]
    if debug:
         print ("   Creating sql query for selecting tables...")
    # get a list with all datainfoids covering the selected time range
    sql = 'SELECT table_name FROM information_schema.tables'
    fulllist = query_db(db,sql,debug=debug)
    if debug:
        print ("got fulllist", fulllist)
    table1list = [el[0] for el in fulllist]
    if debug:
        print ("got tables:", table1list)

    # only select tables with the correct schema
    table2list = [tab for tab in table1list if tab.count("_") == 3]
    if debug:
        print ("got tables:", table2list)

    # remove tables from blacklist
    blacklist.append("INNODB")
    blacklist.append("COLLATION")
    if len(blacklist) > 0:
        table3list = []
        for tab in table2list:
            add = True
            for bl in blacklist:
                if tab.find(bl) > -1:
                    add = False
            if add:
                table3list.append(tab)
    else:
        table3list = table2list

    # select only desired tables
    if len(sensorlist) > 0:
        table4list = []
        for tab in table3list:
            add = False
            for se in sensorlist:
                if tab.find(se) > -1:
                    add = True
            if add:
                table4list.append(tab)
    else:
        table4list = table3list

    if debug:
         print ("   -> tables to be treated: {}".format(table4list))
    return table4list


def main(argv):
    version = "1.0.0"
    conf = ''
    path = ''
    ratio = None
    sensorlist = []
    blacklist = []
    config={}
    sr = np.nan
    startdate = ''
    hostname = socket.gethostname().upper()
    debug=False
    try:
        opts, args = getopt.getopt(argv,"hc:i:s:b:vD",["config=","ratio=","sensorlist=","blacklist=","version=","debug=",])
    except getopt.GetoptError:
        print ('db_truncate.py -c <config> -i <ratio> -s <sensorlist> -b <blacklist> -v <version>')
        sys.exit(2)
    for opt, arg in opts:
        if opt == '-h':
            print ('-------------------------------------')
            print ('Description:')
            print ('-- db_truncate.py removes old data from a databank  --')
            print ('db_truncate.py truncates contents of time series in a MagPy database. ')
            print ('Whereas "archive" also allows for truncating the database (based on DATAINO) ')
            print ('"db_truncate" removes contents from all tables of xxx_xxx_xxxx_xxxx structure. ')
            print ('(independent of DATAINFO contents). ')
            print ('The databank size is automatically restricted ')
            print ('in dependency of the sampling rate of the input data. Only the last ')
            print ('12 days of second data, the last 720 days of minute data and ')
            print ('approximately 118 years of hourly data are kept.')
            print ('-------------------------------------')
            print ('Usage:')
            print ('db_truncate.py -c <config> -i <ratio> -s <sensorlist> -b <blacklist>')
            print ('-------------------------------------')
            print ('Options:')
            print ('-c (required) : provide a path to a configuration file')
            print ('-i            : ratio')
            print ('-b            : blacklist (tables to skip)')
            print ('-s            : sensorlist (use only these tables)')
            print (' given options will override config data')
            print ('-------------------------------------')
            print ('Example:')
            print ('python db_truncate.py -c config')
            sys.exit()
        elif opt in ("-c", "--config"):
            conf = os.path.abspath(arg)
        elif opt in ("-i", "--ratio"):
            ratio = int(arg)
        elif opt in ("-s", "--sensorlist"):
            sensorlist = arg.split(",")
        elif opt in ("-b", "--blacklist"):
            blacklist = arg.split(",")
        elif opt in ("-v", "--version"):
            print ("db_truncate.py version: {}".format(version))
        elif opt in ("-D", "--debug"):
            debug = True

    if debug:
        print ("Running db_truncate.py")
        print ("-------------------------------")

    if os.path.isfile(conf):
        print ("  Read configuration data:")
        config = mm.get_conf(conf)
        print ("   -> configuration data extracted")

    ## Logger configuration data
    logpath = config.get('logpath')
    receiver = config.get('notification')
    receiverconf = config.get('notificationconf')
    ## Analysis configuration data
    cratio = int(config.get('cleanratio',12))
    if not ratio:
        ratio = cratio
    csensorlist = config.get('sensorlist',[])
    if not sensorlist:
        sensorlist = csensorlist
    cblacklist = config.get('blacklist',[])
    if not blacklist:
        blacklist = cblacklist

    if debug:
        print ("Parameters:")
        print ("Ratio:", ratio)
        print ("Sensorlist:", sensorlist)
        print ("Blacklist:", blacklist)

    db = mm.connect_db(config.get('dbcredentials','cobsdb'))

    tables = get_table_tist(db, sensorlist=sensorlist, blacklist=blacklist, debug=debug)

    if debug:
        print ("Cleaning database contents of:", tables)

    for data in tables:
        sr = np.nan
        st = datetime.now(timezone.utc).replace(tzinfo=None)
        amount = DataStream()
        delete = False
        print (" ---------------------------- ")
        print ("Cleaning contents:", data)
        # Test of dataid table exists
        try:
            getline = True
            amount = db.get_lines(data,10000)
        except:
            print ("Could not get lines from data file")
            getline = False

        if getline:
            sr = amount.samplingrate()
            delete = False
            if amount.length()[0] > 0:
                delete = True
                print ("   Deleting old data for {}".format(data))
            else:
                print ("   Doing nothing for table {}".format(data))

        if not isnan(sr) and delete and getline:
            print ("  - Deleting entries in database older than {} days".format(int(sr*ratio)))
            if not debug:
                try:
                    db.delete(data,samplingrateratio=ratio)
                    et = datetime.now(timezone.utc).replace(tzinfo=None)
                    print (" -> ... success: needed {} minutes".format((et-st).total_seconds()/60.))
                except:
                    print (" -> ... failure")
            else:
                print (" DEBUG selected: will not delete anything")
        print (" ---------------------------- ")

if __name__ == "__main__":
   main(sys.argv[1:])



#!/usr/bin/env python
# coding=utf-8

"""
Upload files

DESCRIPTION
   upload data to a destination using various different protocols
   supported are FTP, SFTP, RSYNC, SCP

JOBLIST:
   Jobs are listed in a json structure and read by the upload process.
   You can have multiple jobs. Each job refers to a local path. Each job
   can have multiple destinations.


   {"graphmag" :  {"path":"/home/leon/Tmp/Upload/graph/aut.png",
                    "destinations": {"conradpage": { "type":"ftp", "path" : "/images/graphs/magnetism"} },
                    "log":"/home/leon/Tmp/Upload/testupload.log",
                    "endtime":"utcnow",
                     "starttime":2},
    "wicadjmin" :  {"path":"/home/leon/Tmp/Temp",
                    "destinations": {"gleave": { "type":"sftp", "path" : "/uploads/all-obs"} },
                    "log":"/home/leon/Tmp/wicadjart.log",
                    "endtime":"utcnow",
                     "starttime":2},
   }

## FTP Upload from a directory using files not older than 2 days
{"graphmag" : {"path":"/srv/products/graphs/magnetism/","destinations": {"conradpage": { "type":"ftp", "path" : "images/graphs/magnetism/"} },"log":"/home/leon/Tmp/Upload/testupload.log", "extensions" : ["png"], "namefractions" : ["aut"], "starttime" : 2, "endtime" : "utcnow"}}
## FTP Upload a single file
{"graphmag" : {"path":"/home/leon/Tmp/Upload/graph/aut.png","destinations": {"conradpage": { "type":"ftp", "path" : "images/graphs/magnetism/"} },"log":"/home/leon/Tmp/Upload/testupload.log"}}
## FTP Upload all files with extensions
{"mgraphsmag" : {"path":"/home/leon/Tmp/Upload/graph/","destinations": {"conradpage": { "type":"ftp", "path" : "images/graphs/magnetism/"} },"log":"/home/leon/Tmp/Upload/testupload.log", "extensions" : ["png"]} }
## Test environment
{"TEST" : {"path":"../","destinations": {"homepage": { "type":"test", "path" : "my/remote/path/"} },"log":"/var/log/magpy/testupload.log", "extensions" : ["png"], "starttime" : 2, "endtime" : "utcnow"} }
## RSYNC upload
{"ganymed" : {"path":"/home/leon/Tmp/Upload/graph/","destinations": {"ganymed": { "type":"rsync", "path" : "/home/cobs/Downloads/"} },"log":"/home/leon/Tmp/Upload/testupload.log"} }
## JOB on BROKER
{"magnetsim" : {"path":"/home/cobs/SPACE/graphs/","destinations": {"conradpage": { "type":"ftp", "path" : "images/graphs/magnetism/"} },"log":"/home/cobs/Tmp/testupload.log", "extensions" : ["png"], "namefractions" : ["magvar","gic_prediction","solarwind"], "starttime" : 20, "endtime" : "utcnow"}, "supergrad" : {"path":"/home/cobs/SPACE/graphs/","destinations": {"conradpage": { "type":"ftp", "path" : "images/graphs/magnetism/supergrad"} },"log":"/home/cobs/Tmp/testupload.log", "extensions" : ["png"], "namefractions" : ["supergrad"], "starttime" : 20, "endtime" : "utcnow"},"meteo" : {"path":"/home/cobs/SPACE/graphs/","destinations": {"conradpage": { "type":"ftp", "path" : "images/graphs/meteorology/"} },"log":"/home/cobs/Tmp/testupload.log", "extensions" : ["png"], "namefractions" : ["Meteo"], "starttime" : 20, "endtime" : "utcnow"}, "radon" : {"path":"/home/cobs/SPACE/graphs/","destinations": {"conradpage": { "type":"ftp", "path" : "images/graphs/radon/"} },"log":"/home/cobs/Tmp/testupload.log", "extensions" : ["png"], "namefractions" : ["radon"], "starttime" : 20, "endtime" : "utcnow"}, "title" : {"path":"/home/cobs/SPACE/graphs/","destinations": {"conradpage": { "type":"ftp", "path" : "images/slideshow/"} },"log":"/home/cobs/Tmp/testupload.log", "extensions" : ["png"], "namefractions" : ["title"]}, "gic" : {"path":"/home/cobs/SPACE/graphs/","destinations": {"conradpage": { "type":"ftp", "path" : "images/graphs/spaceweather/gic/"} },"log":"/home/cobs/Tmp/testupload.log", "extensions" : ["png","gif"], "namefractions" : ["24hours"]}, "seismo" : {"path":"/home/cobs/SPACE/graphs/","destinations": {"conradpage": { "type":"ftp", "path" : "images/graphs/seismology/"} },"log":"/home/cobs/Tmp/testupload.log", "extensions" : ["png"], "namefractions" : ["quake"]} }

## JOB on MARTAS2
{"bufferdata" : {"path":"/srv/mqtt/GSM19_7122568_0001/","destinations": {"zamgftp": { "type":"ftp", "path" : "data/beagle/saogabriel"} },"log":"/var/log/magpy/upload.log", "extensions" : ["bin"], "starttime" : 60, "endtime" : "utcnow"}, "env1data" : {"path":"/srv/mqtt/DS18B20_2874EA75D0013CD4_0001/","destinations": {"zamgftp": { "type":"ftp", "path" : "data/beagle/saogabriel"} },"log":"/var/log/magpy/upload.log", "extensions" : ["bin"], "starttime" : 60, "endtime" : "utcnow"}, "env2data" : {"path":"/srv/mqtt/UNOR3_851383136333516112E0_0001/","destinations": {"zamgftp": { "type":"ftp", "path" : "data/beagle/saogabriel"} },"log":"/var/log/magpy/upload.log", "extensions" : ["bin"], "starttime" : 60, "endtime" : "utcnow"}, "backup" : {"path":"/home/debian/Backups/","destinations": {"zamgftp": { "type":"ftp", "path" : "data/beagle/saogabriel"} },"log":"/var/log/magpy/upload.log", "extensions" : ["*"], "starttime" : 60, "endtime" : "utcnow"} }

APPLICTAION:
   python3 file_uploads.py -j /my/path/uploads.json -m /tmp/sendmemory.json
"""


#from magpy.stream import *
import magpy.opt.cred as mpcred

from martas.core.methods import martaslog as ml
from martas.core import methods as mm

import ftplib
from threading import Thread

import getopt
import json
import subprocess
from subprocess import check_output   # used for checking whether send process already finished
from datetime import datetime, timedelta, timezone
import paramiko
import sys, os
from pathlib import Path


# ################################################
#             Methods
# ################################################

# deprecated - replace already by mm.get_json
def getcurrentdata(path):
    """
    usage: getcurrentdata(currentvaluepath)
    example: update kvalue
    fulldict = getcurrentdata(currentvaluepath)
    valdict = fulldict.get('magnetism',{})
    valdict['k'] = [kval,'']
    valdict['k-time'] = [kvaltime,'']
    fulldict[u'magnetism'] = valdict
    writecurrentdata(path, fulldict)
    """
    fulldict = {}
    if os.path.isfile(path):
        with open(path, 'r') as file:
            fulldict = json.load(file)
    else:
        print ("path not found")
    return fulldict

# deprecated - replace already by mm.put_json
def writecurrentdata(path,dic):
    """
    usage: writecurrentdata(currentvaluepath,fulldict)
    example: update kvalue
    see getcurrentdata
    """
    with open(path, 'w', encoding='utf-8') as f:
        json.dump(dic, f, ensure_ascii=False, indent=4)
    return True


def active_pid(name):
     # Part of Magpy starting with version ??
    try:
        pids = map(int,check_output(["pidof",name]).split())
    except:
        return False
    return True


def uploaddata(localpath, destinationpath, typus='ftp', address='', user='', pwd='', port=None, proxy=None, logfile='stdout'):
    """
    DEFINITION:
        upload data method.
        Supports file upload to servers using the following schemes:
        ftp
        sftp     (requires sftp)
        ftpback (background process)
        scp  (please consider using rsync)  scp transfer requires a established key, therefor connect to the server once using ssh to create it
        gin   (curl based ftp upload to data gins)
    """
    success = True
    print ("Running upload to {} (as {}) via {}: {} -> {}, logging to {}".format(address, user, typus, localpath, destinationpath, logfile)) 
    if typus == 'ftpback':
           Thread(target=ftptransfer, kwargs={'source':localpath,'destination':destinationpath,'host':address,'port':port,'user':user,'password':pwd,'logfile':logfile}).start()
    elif typus == 'ftp':
           success = ftptransfer(source=localpath,destination=destinationpath,host=address,port=port,user=user,password=pwd,logfile=logfile)
    elif typus == 'sftp':
           success = sftptransfer(source=localpath,destination=destinationpath,host=address,user=user,password=pwd,proxy=proxy,logfile=logfile)
    elif typus == 'scp':
           timeout = 60
           destina = "{}@{}:{}".format(user,address,destinationpath)
           mm.scptransfer(localpath,destina,pwd,timeout=timeout)
    elif typus == 'rsync':
           # create a command line string with rsync ### please note,,, rsync requires password less comminuctaion
           rsyncstring = "rsync -avz -e ssh {} {}".format(localpath, user+'@'+address+':'+destinationpath)
           print ("Executing:", rsyncstring)
           subprocess.call(rsyncstring.split())
    elif typus.startswith('gin'):
        if not active_pid('curl'):
            print ("  -- Uploading minute data to GIN - active now")
            stdout = False
            if logfile == 'stdout':
                stdout = True
            print (" -- calling GINUPLOAD")
            stdout = True
            if typus.endswith('basic'):
                success = ginupload(localpath, user, pwd, address, authentication=' ',stdout=stdout)
            else:
                success = ginupload(localpath, user, pwd, address, stdout=stdout)
            print ("   -> Done GINUPLOAD")
        else:
            print ("curl is active")
    elif typus == 'test':
           print ("No file transfer - just a test run")
    else:
        print ("Selected type of transfer is not supported")

    return success


def getchangedfiles(basepath,memory,startdate=datetime(1777,4,30),enddate=None, extensions=None, namefractions=None, add="newer", debug=False):
    """
    DESCRIPTION
        Will compare contents of basepath and memory and create a list of paths with changed information
        This method will work a bit like rsync without accessing the zieldirectory. It just checks whether
        such data has been uploaded already
    VARIABLES
        basepath  (String)      :  contains the basepath in which all files will be checked
        memory    (list/dict)   :  a dictionary with filepath and last change date (use getcurrentdata)
        startdate (datetime)    :  changes after startdate will be considered
        enddate   (datetime)    :  changes before enddate will be considered
        add       (string)      :  either "all" or "newer" (default)
    RETURNS
        dict1, dict2   : dict1 contains all new data sets to be uploaded, dict2 all analyzed data files for storage
    """
    if not enddate:
        enddate = datetime.now(timezone.utc).replace(tzinfo=None)
    if not extensions:
        extensions = []
    if not namefractions:
        namefractions = []
    startdate = startdate.replace(tzinfo=None)
    enddate = enddate.replace(tzinfo=None)


    filelist=[]
    try:
        if os.path.isfile(basepath):
            filelist = [basepath]
        else:
            for file in os.listdir(basepath):
                fullpath=os.path.join(basepath, file)
                filename, file_extension = os.path.splitext(fullpath)
                if os.path.isfile(fullpath):
                    if isinstance(extensions,list) and len(extensions) > 0 and any([file_extension.find(ext)>-1 for ext in extensions]) and len(namefractions) > 0:
                        if isinstance(namefractions,list) and len(namefractions) > 0 and any([filename.find(frac)>-1 for frac in namefractions]):
                            filelist.append(fullpath)
                    elif isinstance(extensions,list) and len(extensions) > 0 and any([file_extension.find(ext)>-1 for ext in extensions]) and len(namefractions) == 0:
                        filelist.append(fullpath)
                    elif isinstance(namefractions,list) and len(namefractions) > 0 and any([filename.find(frac)>-1 for frac in namefractions]) and len(extensions) == 0:
                        filelist.append(fullpath)
                    elif len(extensions) == 0 and len(namefractions) == 0:
                        filelist.append(fullpath)
    except:
        print ("Directory not found")
        return {}, {}

    if debug:
        print ("getchangedfiles 1:", filelist, startdate, enddate)


    retrievedSet={}
    for name in filelist:
        mtime = datetime.fromtimestamp(os.path.getmtime(name))
        stat=os.stat(os.path.join(basepath, name))
        mtime=stat.st_mtime
        if datetime.fromtimestamp(mtime, timezone.utc).replace(tzinfo=None) > startdate and datetime.fromtimestamp(mtime, timezone.utc).replace(tzinfo=None) <= enddate:
            retrievedSet[name] = mtime

    if debug:
        print ("getchangedfiles 2:", retrievedSet)
        print ("getchangedfiles 3:", memory)
    if memory:
        if sys.version_info >= (3,):
            newdict = dict(retrievedSet.items() - memory.items())
        else:
            newdict = dict(filter(lambda x: x not in memory.items(), retrievedSet.items()))
    else:
        newdict = retrievedSet.copy()

    return newdict, retrievedSet


def sftptransfer(source, destination, host="yourserverdomainorip.com", user="root", password="12345", port=22, proxy=None, logfile='stdout'):
    """
    DEFINITION:
        Tranfering data to an sftp server
        (proxy support not tested - please additionally use corkscrew for tranfer through a proxy)

    PARAMETERS:
    Variables:
        - source:       (str) Path within ftp to send file to.
        - destination:  (str) full file path to send to.
        - host:         (str) address of reciever
        - user:
        - password:
        - proxy         (tuple) like ("123.123.123.123",8080)
        - logfile:      (str) not used so far


    RETURNS:
        - success (BOOL) True if succesful.

    EXAMPLE:
        sftptransfer(source='/home/me/file.txt', destination='/data/magnetism/file.txt', host='www.example.com', user='mylogin', password='mypassword', logfile='/home/me/Logs/magpy-transfer.log'
                            )
    """

    if not os.path.isfile(source):
        print ("source does not exist ... aborting")
        return False

    #if not logfile == 'stdout':
    #    paramiko.util.log_to_file(logfile)

    if not proxy:
        #transport = paramiko.Transport((host, port))
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        try:
            ssh.connect(host, username=user, password=password, timeout=10)
            with ssh.open_sftp() as sftp:
                destina = os.path.join(destination,os.path.basename(source))
                sftp.put(source, destina)
        except:
            print("failed to connect client ... aborting")
            return False
    else:
        print ("Using proxy (needs to a tuple (paddr,pport)): {}".format(proxy))  
        import socks
        s = socks.socksocket()
        s.set_proxy(
              proxy_type=socks.SOCKS5,
              addr=proxy[0],
              port=proxy[1]
        )
        try:
            s.connect((host,port))
            transport = paramiko.Transport(s)
            transport.connect(username=user,password=password)
            with paramiko.SFTPClient.from_transport(transport) as client:
                destina = os.path.join(destination,os.path.basename(source))
                client.put(source, destina)
        except:
            print("failed to connect client ... aborting")
            return False

    return True


def ftptransfer (source, destination, host="yourserverdomainorip.com", user="root", password="12345", port=22, proxy=None, logfile='stdout', cleanup=False, debug=False):
    """
    DEFINITION:
        Tranfering data to an ftp server

    PARAMETERS:
    Variables:
        - source:       (str) Path within ftp to send file to.
        - destination:  (str) full file path to send to.
        - host:         (str) address of reciever
        - user:
        - password:
        - proxy         (tuple) like ("123.123.123.123",8080)
        - logfile:      (str) not used so far
        - cleanup:      (bool) If True, transfered files are removed from the local directory.

    RETURNS:
        - True/False.

    EXAMPLE:
        ftpdatatransfer(
                localfile='/home/me/file.txt',
                ftppath='/data/magnetism/this',
                myproxy='www.example.com',
                login='mylogin',
                passwd='mypassword',
                logfile='/home/me/Logs/magpy-transfer.log'
                            )

    APPLICATION:
    """

    if not source:
        return False
    else:
        try:
            filename = os.path.split(source)[1]
        except:
            print ("Could not determine filename of source")
            return False

    if debug:
        print ("ftptransfer: running ftp transfer")

    def ftpjob(site,source,destination,filename):
        site.set_debuglevel(1)
        site.cwd(destination)
        try:
            site.delete(filename)
        except:
            print ("File not yet existing")
        if debug:
            print ("Uploading file ... {}".format(filename))
        try:
            with open(source, 'rb') as image_file:
                site.storbinary('STOR {}'.format(filename), image_file)
        except:
            image_file = open(source, 'rb')
            site.storbinary('STOR {}'.format(filename), image_file)
            image_file.close()


    transfersuccess = False
    tlsfailed = False
    # First try to connect via TLS
    try:
        with ftplib.FTP_TLS(host, user, password) as ftp:
            print ("FTP TLS connection established...")
            ftpjob(ftp,source,destination,filename)
            transfersuccess = True
    except:
        tlsfailed = True

    # First try to connect without TLS
    if tlsfailed:
        try:
            with ftplib.FTP(host, user, password) as ftp:
                print ("FTP connection established...")
                ftpjob(ftp,source,destination,filename)
                transfersuccess = True
        except:
            try:
                # fallback without with (seems to be problemtic on some old py2 machines
                ftp = ftplib.FTP(host, user, password)
                print ("FTP connection established...")
                ftpjob(ftp,source,destination,filename)
                ftp.close()
                transfersuccess = True                
            except:
                pass

    if cleanup and transfersuccess:
        os.remove(source)

    return transfersuccess

def ginupload(filename, user, password, url, authentication=' --digest ', stdout=True):
    """
    DEFINITION:
        Method to upload data to the Intermagnet GINs using curl
        tries to upload data to the gin - if not succesful ...
    PARAMETERS:
        filename        (string) filename including path
        user            (string) GIN user
        password        (string) GIN passwd
        url             (string) url address (e.g. http://app.geomag.bgs.ac.uk/GINFileUpload/Cache)
        stdout          (bool) if True the return will be printed
    REQUIRES:
        - a working installation of curl
        - the package subprocess
    """
    print ("  Starting ginupload")
    print ("  ------------------")
    faillog = ''
    logpath = ''
    success = False

    if not logpath:
        logpath = "/tmp/ginupload.log"

    commandlist = []

    print ("Running ginupload")

    curlstring = 'curl -F "File=@'+filename+';type=text/plain" -F "Format=plain" -F "Request=Upload" -u '+user+':'+password+authentication+url
    print ("CURL command looks like {}".format(curlstring))
    if not curlstring in commandlist:
        commandlist.append(curlstring)

    # Process each upload
    # -----------------------------------------------------
    for command in commandlist:
        print ("Running command:", command)
        p = subprocess.Popen(command,stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        out, error = p.communicate()
        print (" -> Done")
        if sys.version_info.major >= 3:
            out = out.decode('ascii')
            error = error.decode('ascii')
        if stdout:
            print(out, error)
        if "Success" in out:
            success = True

    return success


def main(argv):
    statusmsg = {}
    jobs = ''
    tele = ''
    sendlogpath=''
    logpath = ''
    name = ''
    receiver = ''
    receiverconf = ''
    proxies = {}
    debug = False
    try:
        opts, args = getopt.getopt(argv,"hj:m:t:",["jobs=","memory=","telegram=",])
    except getopt.GetoptError:
        print ('file_upload.py -j <jobs> -m <memory> -t <telegram>')
        sys.exit(2)
    for opt, arg in opts:
        if opt == '-h':
            print ('-------------------------------------')
            print ('Description:')
            print ('-- file_uploads.py send data to any destination of your choice  --')
            print ('-----------------------------------------------------------------')
            print ('file_uploads is a python wrapper allowing to send either by')
            print ('ftp, scp, sftp using a similar inputs.')
            print ('It can handle several different uploads at once.')
            print ('Upload parameters have to be provided in a json structure.')
            print ('file_uploads requires magpy >= 2.0.0.')
            print ('-------------------------------------')
            print ('Usage:')
            print ('file_uploads.py -j <jobs> -m <memory> -t <telegram>')
            print ('-------------------------------------')
            print ('Options:')
            print ('-j (required) : a json structure defining the uploads')
            print ('-m (required) : a path for "memory"')
            print ('-------------------------------------')
            print ('Example of jobs structure:')
            print ('{"wicadjmin":{"path":"/home/leon/Tmp/Temp","destinations":{"gleave":{"type":"sftp", "path" : "/uploads/all-obs"}},')
            print ('"log":"/home/leon/Tmp/wicadjart.log","endtime":"utcnow","starttime":2}}')
            print ('! please note endtime is usually utcnow')
            print ('! starttime is given as a timedelta in days toward endtime, an integer is required')
            print ('-------------------------------------')
            print ('Application:')
            print ('python3 file_uploads.py -j /my/path/uploads.json -m /tmp/sendmemory.json')
            sys.exit()
        elif opt in ("-j", "--jobs"):
            jobs = arg
        elif opt in ("-m", "--memory"):
            sendlogpath = arg
        elif opt in ("-t", "--telegram"):
            tele = arg
        elif opt in ("-D", "--debug"):
            debug = True

    if tele:
        # ################################################
        #          Telegram Logging
        # ################################################
        # tele needs a logpath, and a configuration path ('/home/cobs/SCRIPTS/telegram_notify.conf')
        mainpath = os.path.dirname(os.path.realpath(__file__))
        logpath = os.path.join(mainpath, "..", "log", "mm-fu-uploads.log")
        configpath = os.path.join(mainpath, "..", "conf", "monitor.cfg")
        if os.path.isfile(configpath):
            print (" reading configuration file ...")
            monitorconf = mm.get_conf(configpath)
            receiver = monitorconf.get('notification')
            receiverconf = monitorconf.get('notificationconf')
            if monitorconf.get('https'):
                proxies['https'] = monitorconf.get('https')
            if monitorconf.get('http'):
                proxies['http'] = monitorconf.get('http')

    if jobs == '':
        print ('Specify a valid path to a jobs dictionary (json):')
        print ('-- check file_uploads.py -h for more options and requirements')
        sys.exit()
    else:
        if os.path.isfile(jobs):
            workdictionary = mm.get_json(jobs)
            #mm.put_json("/home/leon/.martas/conf/fileuploadjobs.json", workdictionary)
        else:
            print ('Specify a valid path to a jobs dictionary (json):')
            print ('-- check file_uploads.py -h for more options and requirements')
            sys.exit()

    # make fileupload an independent method importing workingdictionary (add to MARTAS?)
    if not sendlogpath:
        sendlogpath = '/tmp/lastupload.json'
    else:
        # create a new memory if not existing
        if not os.path.isfile(sendlogpath):
            filepath = Path(sendlogpath)
            filepath.parent.mkdir(parents=True, exist_ok=True)

    """
    Main Prog
    """
    try:
        for key in workdictionary:
            name = "FileUploads-{}".format(key)
            print("----------------------------------------------")
            print ("DEALING with ", key)
            print("----------------------------------------------")
            lastfiles = {}
            fulldict = {}
            if os.path.isfile(sendlogpath):
                if os.stat(sendlogpath).st_size == 0:
                    lastfiles = {}
                else:
                    with open(sendlogpath, 'r') as file:
                        fulldict = json.load(file)
                        lastfiles = fulldict.get(key)
                        # lastfiles looks like: {'/path/to/my/file81698.txt' : '2019-01-01T12:33:12', ...}

            if not lastfiles == {}:
                print ("creating new memory")
                pass

            #print ("Extracting jobs ...")
            sourcepath = workdictionary.get(key).get('path')
            extensions = workdictionary.get(key).get('extensions',[])
            namefractions = workdictionary.get(key).get('namefractions',[])
            # Test if sourcepath is file
            starttime = workdictionary.get(key).get('starttime',datetime(1777,4,30))
            now = datetime.now(timezone.utc).replace(tzinfo=None)
            endtime = workdictionary.get(key).get('endtime',now)
            if endtime in ["utc","now","utcnow"]:
                endtime = now
            if isinstance(starttime, int):
                starttime = now-timedelta(days=starttime)
            #print ("Done ...")
            newfiledict, alldic = getchangedfiles(sourcepath, lastfiles, starttime, endtime, extensions,namefractions)

            #print ("Found new: {} and all {}".format(newfiledict, alldic))

            for dest in workdictionary.get(key).get('destinations'):
                print ("  --------------------------------------------")
                print ("  -> Destination: {}".format(dest))
                print ("  --------------------------------------------")
                print ("")
                address=mpcred.lc(dest,'address')
                user=mpcred.lc(dest,'user')
                passwd=mpcred.lc(dest,'passwd')
                port=mpcred.lc(dest,'port')
                destdict = workdictionary.get(key).get('destinations')[dest]
                proxy = destdict.get('proxy',None)
                #print (destdict)
                if address and user and newfiledict:
                    for nfile in newfiledict:
                        print ("    -> Uploading {} to dest {}".format(nfile, dest))
                        success = uploaddata(nfile, destdict.get('path'), destdict.get('type'), address, user, passwd, port, proxy=proxy, logfile=destdict.get('logfile','stdout'))
                        print ("    -> Success", success)
                        if not success:
                            #remove nfile from alldic
                            # thus it will be retried again next time
                            print (" !---> upload of {} not successful: keeping it in todo list".format(nfile))
                            del alldic[nfile]

            print (" ... done for key", key)
            fulldict[key] = alldic
            mm.put_json(sendlogpath, fulldict)
            print(" ... log written")
        statusmsg[name] = "uploading data successful"
    except:
        statusmsg[name] = "error when uploading files - please check"

    if tele:
        if debug:
            print(statusmsg)
        elif receiver:
            martaslog = ml(logfile=logpath, receiver=receiver, proxies=proxies)
            if receiver == 'email':
                martaslog.email['config'] = receiverconf
            elif receiver == 'telegram':
                martaslog.telegram['config'] = receiverconf
            changes = martaslog.msg(statusmsg)

    print ("----------------------------------------------------------------")
    print ("file upload app finished")
    print ("----------------------------------------------------------------")
    print ("SUCCESS")


if __name__ == "__main__":
   main(sys.argv[1:])






"""
cr1000jcprotocol written 2017 by richard.mandl@zamg.ac.at
for MQTT protocol by Roman Leonhardt and Rachel Bailey to be used in the Conrad Observatory.
Makes sense only when using a CR1000 data logger with specific set-up
in combination with a Judd JC depth sensor 
to measure snow depth
"""

# ###################################################################
# Import packages
# ###################################################################

import struct # for binary representation
from datetime import datetime, timedelta, timezone
from twisted.python import log
from martas.core import methods as mm
import threading
import time

try:
    from pycampbellcr1000 import CR1000
    cr1000imported = True
except:
    log.msg("pycampbellcr1000 package not available")
    cr1000imported = False

# ###################################################################

# some helpful function
def datetime2array(t):
    return [t.year,t.month,t.day,t.hour,t.minute,t.second,t.microsecond]

# sensor height in cm
# TODO 184 ist fuer die Hohe Warte am Schreibtisch
SENSOR_HEIGHT = 184

if cr1000imported:
    class cr1000jcProtocol:        
        """
        Protocol to read from a campbell datalogger CR1000
        The CR1000's tables are configured to log Judd JC data
        """

        device=None
        reconnect=threading.Event()

        def __init__(self, client, sensordict, confdict):
            self.client = client
            self.sensordict = sensordict 
            self.confdict = confdict
            # variables for broadcasting via mqtt:
            self.count=0
            self.datalst = []
            #self.datacnt = 0
            self.metacnt = 10
            ###
            port = confdict['serialport']+sensordict.get('port')
            baudrate = sensordict.get('baudrate')
            # string should look like that: device = CR1000.from_url('serial:/dev/ttyUSB3:38400')
            #try: 
            log.msg('connecting to device...')
            self.device = CR1000.from_url('serial:{}:{}'.format(port,baudrate))
            #except:
                # TODO no CR1000 present... "Can not access to device"
                #log.msg('Can not access to device')
            tables = self.device.list_tables()
            if not tables == ['Status', 'SamplesEvery2s', 'ValuesEveryMinute', 'Public']:
                # TODO log-Befehl, Abbruch!
                log.msg('CR1000 not configured for Judd JC')
            else:
                self.device.settime(datetime.now(timezone.utc).replace(tzinfo=None))
                # TODO folgende Ausgabe nur fuers Programmieren!!
                log.msg('++++++++ Information von den Konfig-Dateien +++++++++++++')
                log.msg('++ client:')
                log.msg(client)
                log.msg('++ sensordict:')
                log.msg(sensordict)
                log.msg('++ confdict:')
                log.msg(confdict)

        def sendRequest(self):
            # TODO wohin mit debug?
            debug = False
            if self.reconnect.is_set():
                log.msg('exiting, mutex locked!')
                return
            t = datetime.now(timezone.utc).replace(tzinfo=None)
            past = t-timedelta(seconds=3)
            vals = self.device.get_data('SamplesEvery2s',past,t)
            # vals[0] because we grap no older data, there is only one value in 2 seconds
            # timestamp directly from datetime into array
            # TODO Roman fragen, ob oder wie Vergleich mit Computerzeit
            try:
                darray = datetime2array(vals[0]['Datetime'])
                # TODO "again" ist Provisorium
                again = False
            except:
                again = True
            try:
                if again:
                    t = datetime.now(timezone.utc).replace(tzinfo=None)
                    past = t-timedelta(seconds=3)
                    darray = datetime2array(vals[0]['Datetime'])
                    log.msg("IT TOOK A SECOND TIME TO GET DATA PROPERLY!") 
            except:
                # there will be no log messages when the logger is turned off
                return
                # TODO reconnect Loesung, nur wenn sie sauber funktioniert!
                log.msg('NO DATA FROM CR1000 !!! - vals:')
                log.msg(vals)
                port = self.confdict['serialport']+self.sensordict.get('port')
                baudrate = self.sensordict.get('baudrate')
                self.reconnect.set()
                connected=False
                while not connected:
                    self.device.bye()
                    log.msg('reconnecting to device...')
                    time.sleep(5)
                    try:
                        self.device = CR1000.from_url('serial:{}:{}'.format(port,baudrate))
                        tables = self.device.list_tables()
                        if tables == ['Status', 'SamplesEvery2s', 'ValuesEveryMinute', 'Public']:
                            connected=True
                            log.msg('schaut ok aus...')
                            time.sleep(2)
                    except:
                        log.msg('reconnect wohl missglueckt!')
                    try:
                        past = t-timedelta(seconds=3)
                        vals = self.device.get_data('SamplesEvery2s',past,t,debug=True)
                        log.msg(SENSOR_HEIGHT*1000.-vals[0]['DiffVolt']*250.)
                    except:
                        log.msg('...wohl doch nicht!')
                        connected=False
                self.reconnect.clear()
                log.msg('mutex released...')
                return
            if debug:
                log.msg('getting data...')
            # snowheight (1000mV is 250cm) - values from CR1000 in mV, factor 1000 for packing
            snowheight = SENSOR_HEIGHT*1000.-vals[0]['DiffVolt']*250.
            darray.append(int(round(snowheight)))
            # TODO weg
            if debug:
                log.msg(darray)

            # preparations for file save
            # date 6 short microsecond unsigned long snowheight signed long
            # TODO alter packcode!
            packcode = "6hLl"
            #packcode = "<6hLl"
            # header 
            sensorid = self.sensordict['sensorid']
            header = "# MagPyBin %s %s %s %s %s %s %d" % (sensorid,'[f]','[JC]','[cm]','[1000]',packcode,struct.calcsize(packcode))
            data_bin = struct.pack(packcode,*darray)
            # date of dataloggers timestamp
            filedate = datetime.strftime(datetime(darray[0],darray[1],darray[2]), "%Y-%m-%d")
            if not self.confdict.get('bufferdirectory','') == '':
                mm.data_to_file(self.confdict.get('bufferdirectory'), sensorid, filedate, data_bin, header)
                if debug:
                    log.msg('Daten gesichert...')

            # sending via MQTT
            data = ','.join(list(map(str,darray)))
            head = header
            topic = self.confdict.get('station') + '/' + self.sensordict.get('sensorid')
            coll = int(self.sensordict.get('stack'))
            if coll > 1:
                self.metacnt = 1 # send meta data with every block
                if self.datacnt < coll:
                    self.datalst.append(data)
                    self.datacnt += 1
                else:
                    senddata = True
                    data = ';'.join(self.datalst)
                    self.datalst = []
                    self.datacnt = 0
            else:
                senddata = True

            if senddata:
                self.client.publish(topic+"/data", data)
                if self.count == 0:
                    self.client.publish(topic+"/meta", head)
                self.count += 1
                if self.count >= self.metacnt:
                    self.count = 0


            
            # right now auxiliary data only in the log file
            if t.second<2:
                # every minute aux data (battery voltage and logger temperature) will be available
                # going 61s into the past to make sure there are already data
                past=t-timedelta(seconds=61)
                aux=self.device.get_data('ValuesEveryMinute',past,t)
                log.msg('----- aux every minute:')
                # timestamp directly from datetime into array
                try:
                    darray = datetime2array(aux[0]['Datetime'])
                except:
                    # following should never happen...
                    log.msg('AUXILIARY DATA NOT GOT PROPERLY! - aux:')
                    log.msg(aux)
                    log.msg('trying again...')
                    past=t-timedelta(seconds=62)
                    aux=self.device.get_data('ValuesEveryMinute',past,t)
                    try:
                        darray = datetime2array(aux[0]['Datetime'])
                    except:
                        log.msg('giving up...')
                        return

                # battery voltage - factor 1000 for packing
                BattV_Min=int(round(aux[0]['BattV_Min']*1000))
                PTemp_C_Avg=int(round(aux[0]['PTemp_C_Avg']*1000))
                darray.extend([BattV_Min,PTemp_C_Avg])
                # alernative reading:
                #aux = (aux.filter(('Datetime', 'BattV_Min','PTemp_C_Avg')).to_csv(header=False))
                log.msg(darray)
                packcode="<6hLLl"




try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

import sys
import shutil
exec(open('martas/version.py').read())
shutil.copyfile('martas/acquisition.py','scripts/acquisition')
shutil.copyfile('martas/collector.py','scripts/collector')
shutil.copyfile('martas/martas_init.py', 'scripts/martas_init')
shutil.copyfile('martas/web/martas_view.py', 'scripts/martas_view')
shutil.copyfile('martas/web/martas_view.py', 'scripts/marcos_view')

# modules:
# paramiko : file_upload

install_requires=[
            "geomagpy > 1.1.9",
            "numpy >= 1.21.0",
            "scipy >= 1.7.1",
            "pexpect",
            "paramiko",
            "pyserial",
            "requests",
            "python-crontab",
            "twisted",
            "plotly",
            "dash",
            "dash_daq",
            "dash-bootstrap-components",
            "psutil",
            "setuptools >= 61.0.0"
          ]

setup(
    name='martas',
    version=__version__,
    author='R. Leonhardt, R. Bailey, R. Mandl, V. Haberle',
    author_email='roman.leonhardt@geosphere.at',
    packages=['martas', 'martas.app', 'martas.conf', 'martas.core', 'martas.doc', 'martas.init', 'martas.lib', 'martas.logrotate', 'martas.scripts','martas.web', 'martas.web.assets', 'martas.web.pages'],
    scripts=['scripts/collector','scripts/acquisition', 'scripts/martas_init', 'scripts/martas_view', 'scripts/marcos_view'],
    url='',
    license='LICENSE.txt',
    description='MARTAS',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    package_data={'martas': ['conf/*.cfg', 'conf/*.json', 'doc/*.png', 'doc/*.pdf', 'doc/*.md', 'init/*.sh', 'init/*.json', 'install/*.sh', 'logrotate/*.logrotate', 'scripts/*.sh', 'web/*.js', 'web/*.html', 'web/*.py', 'web/assets/*', 'web/pages/*']  },
    install_requires=install_requires,
)



#!/usr/bin/env python
"""
MQTT collector routine of MARCOS:
MQTT protocol as used at the Conrad Observatory.
written by Roman Leonhardt
"""
# ###################################################################
# Import packages
# ###################################################################

## Import MagPy
## -----------------------------------------------------------

from magpy.stream import DataStream, join_streams, KEYLIST, NUMKEYLIST, subtract_streams
from magpy.core import database
from magpy.opt import cred as mpcred

## Import Twisted for websocket and logging functionality
from twisted.python import log
from twisted.web.server import Site
from twisted.web.static import File
from twisted.internet import reactor

import threading
from multiprocessing import Process
import struct
from datetime import datetime
from matplotlib.dates import num2date
import numpy as np
import json
import socket
from io import StringIO

## Import specific MARTAS packages
## -----------------------------------------------------------
from martas.core import methods as mm
from martas.version import __version__
from martas.core.methods import martaslog as ml
from martas.core.websocket_server import WebsocketServer

## Import MQTT
## -----------------------------------------------------------
import paho.mqtt
import paho.mqtt.client as mqtt
import sys, getopt, os
import ssl
try:
    from sslpsk2.sslpsk2 import _ssl_set_psk_server_callback, _ssl_set_psk_client_callback
except:
    pass


## Python Version
## -----------------------------------------------------------
import platform
pyversion = platform.python_version()

## Import Webserver for displaying data
## -----------------------------------------------------------
ws_available = True

# Some variable initialization
## -----------------------------------------------------------
#global counter  # use for diffcalc

qos = 0
streamdict = {}
stream = DataStream()
diffstruct = {}
st = []
senslst = []
headdict = {} # store headerlines for all sensors (headerline are firstline for BIN files)
headstream = {}
verifiedlocation = False
destination = 'stdout'
location = '/tmp'
credentials = 'cred'
stationid = 'WIC'
stid = stationid
webpath = './web'
webport = 8080
socketport = 5000
diffsens = "G823"
blacklist = []
counter = 0
diffcounter = 0
prevlasttimestamp_prim = None

## SSL PSK tools
def _ssl_setup_psk_callbacks(sslobj):
    psk = sslobj.context.psk
    hint = sslobj.context.hint
    identity = sslobj.context.identity
    if psk:
        if sslobj.server_side:
            cb = psk if callable(psk) else lambda _identity: psk
            _ssl_set_psk_server_callback(sslobj, cb, hint)
        else:
            cb = psk if callable(psk) else lambda _hint: psk if isinstance(psk, tuple) else (psk, identity)
            _ssl_set_psk_client_callback(sslobj, cb)


class SSLPSKContext(ssl.SSLContext):
    @property
    def psk(self):
        return getattr(self, "_psk", None)

    @psk.setter
    def psk(self, psk):
        self._psk = psk

    @property
    def hint(self):
        return getattr(self, "_hint", None)

    @hint.setter
    def hint(self, hint):
        self._hint = hint

    @property
    def identity(self):
        return getattr(self, "_identity", None)

    @identity.setter
    def identity(self, identity):
        self._identity = identity


class SSLPSKObject(ssl.SSLObject):
    def do_handshake(self, *args, **kwargs):
        if not hasattr(self, '_did_psk_setup'):
            _ssl_setup_psk_callbacks(self)
            self._did_psk_setup = True
        super().do_handshake(*args, **kwargs)


class SSLPSKSocket(ssl.SSLSocket):
    def do_handshake(self, *args, **kwargs):
        if not hasattr(self, '_did_psk_setup'):
            _ssl_setup_psk_callbacks(self)
            self._did_psk_setup = True
        super().do_handshake(*args, **kwargs)


SSLPSKContext.sslobject_class = SSLPSKObject
SSLPSKContext.sslsocket_class = SSLPSKSocket


class protocolparameter(object):
    def __init__(self):
        self.identifier = {}
        #self.counter = 0

po = protocolparameter()

## WebServer Methods
## -----------------------------------------------------------
def wsThread(wsserver):
    wsserver.set_fn_new_client(new_wsclient)
    wsserver.set_fn_message_received(message_received)
    wsserver.run_forever()

if ws_available:
    global wsserver

# TODO: find a way how to use these two functions:
def new_wsclient(ws_client,server):
    pass
    # for debug: see which threads are running:
    #print(str(threading.enumerate()))

def message_received(ws_client,server,message):
    pass

def webProcess(webpath,webport):
    """
    These few lines will be started as an own process.
    When the main process is killed, also this child process is killed
    because of having it started as a daemon
    """
    resource = File(webpath)
    print ("TESTING", webpath, resource)
    factory = Site(resource)
    #endpoint = endpoints.TCP4ServerEndpoint(reactor, 8888)
    #endpoint.listen(factory)
    # args! -> make an integer out of it
    webport = int(webport)
    reactor.listenTCP(webport,factory)
    reactor.run()

def connectclient(broker='localhost', port=1883, timeout=60, credentials='', user='', password='', qos=0, mqttcert="", mqttpsk="", mqttversion=2, destinationid='',debug=False):
        """
    connectclient method
    used to connect to a specific client as defined by the input variables
    eventually add multiple client -> {"clients":[{"broker":"192.168.178.42","port":"1883"}]} # json type
                import json
                altbro = json.loads(altbrocker)
        """
        ## create a unique clientid consisting of broker, client and destination
        client = None
        hostname = socket.gethostname()
        clientid = "{}{}{}".format(broker,hostname,destinationid)

        ## create MQTT client
        ##  ----------------------------
        pahovers = paho.mqtt.__version__
        pahomajor = int(pahovers[0])
        print(" paho-mqtt version ", pahovers)
        try:
            client = mqtt.Client(clientid, False)
        except:
            try:
                client = mqtt.Client(mqtt.CallbackAPIVersion.VERSION2, client_id=clientid)
            except:
                client = mqtt.Client(mqtt.CallbackAPIVersion.VERSION1, client_id=clientid)

        # TLS encryption part
        if int(port) == 8883 and not mqttpsk:
            if mqttcert:
                if debug:
                    print("MQTT: TLS encryption based on certificate")
                    print(mqttcert)
                client.tls_set(ca_certs=mqttcert)
            else:
                if debug:
                    print("MQTT: basic TLS")
                client.tls_set(ca_certs=None, certfile=None, keyfile=None, cert_reqs=ssl.CERT_REQUIRED,
                               tls_version=ssl.PROTOCOL_TLS,
                               ciphers=None)
        if int(port) in [8884, 8883] and mqttpsk:
            if debug:
                print("MQTT: TLS encrytion based in PSK")
            # making use of discussions in https://github.com/eclipse-paho/paho.mqtt.python/issues/451
            pskidentity = mpcred.lc(mqttpsk, 'user')
            pskpwd = mpcred.lc(mqttpsk, 'passwd')
            #context = SSLPSKContext(ssl.PROTOCOL_TLS_CLIENT) # This does bot work for beaglebone (python3.11 clients)
            print ("MARTAS: Deprecation warning - but necessary for old clients")
            context = SSLPSKContext(ssl.PROTOCOL_TLSv1_2)
            context.set_ciphers('PSK')
            context.psk = bytes.fromhex(pskpwd)
            context.identity = pskidentity.encode()
            client.tls_set_context(context)  # Here we apply the new `SSLPSKContext`

        # Authentication part
        if not credentials in ['','-']:
            # use user and pwd from credential data if not yet set
            if user in ['',None,'None','-']:
                user = mpcred.lc(credentials,'user')
            if password  in ['','-']:
                password = mpcred.lc(credentials,'passwd')
        if not user in ['',None,'None','-']:
            #client.tls_set(tlspath)  # check http://www.steves-internet-guide.com/mosquitto-tls/
            client.username_pw_set(user, password=password)  # defined on broker by mosquitto_passwd -c passwordfile user
        client.on_connect = on_connect
        # on message needs: stationid, destination, location
        client.on_message = on_message
        client.connect(broker, port, timeout)

        return client


def analyse_meta(header,sensorid,debug=False):
    """
    source:mqtt:
    Interprete header information
    """
    if debug:
        print ("Analyzing {}: {}".format(sensorid,header))
    if pyversion.startswith('2'):
        header = header.decode('utf-8')
    # some cleaning actions for false header inputs
    header = header.replace(', ',',')
    header = header.replace('deg C','deg')
    header = header.replace('T (out','T(out')
    h_elem = header.strip().split()
    if not h_elem[-2].startswith('<'): # e.g. LEMI
        packstr = '<'+h_elem[-2]+'B'
    else:
        packstr = h_elem[-2]
    packstr = packstr.encode('ascii','ignore')
    lengthcode = struct.calcsize(packstr)
    si = h_elem[2]
    if not si == sensorid:
        log.msg("Different sensorids in publish address {} and header {} - please check - aborting".format(sensorid,si))
        sys.exit()
    keylist = h_elem[3].strip('[').strip(']').split(',')
    elemlist = h_elem[4].strip('[').strip(']').split(',')
    unitlist = h_elem[5].strip('[').strip(']').split(',')
    multilist = list(map(float,h_elem[6].strip('[').strip(']').split(',')))
    if debug:
        log.msg("Packing code: {}".format(packstr))
        log.msg("keylist: {}".format(keylist))
    po.identifier[sensorid+':packingcode'] = packstr
    po.identifier[sensorid+':keylist'] = keylist
    po.identifier[sensorid+':elemlist'] = elemlist
    po.identifier[sensorid+':unitlist'] = unitlist
    po.identifier[sensorid+':multilist'] = multilist


def create_head_dict(header,sensorid):
    """
    source:mqtt:
    Interprete header information
    """
    head_dict={}
    try:
        # python2.7
        header = header.decode('utf-8')
    except:
        pass
    # some cleaning actions for false header inputs
    header = header.replace(', ',',')
    header = header.replace('deg C','deg')
    h_elem = header.strip().split()
    if not h_elem[-2].startswith('<'):
        packstr = '<'+h_elem[-2]+'B'
    else: # LEMI
        packstr = h_elem[-2]
    packstr = packstr.encode('ascii','ignore')
    lengthcode = struct.calcsize(packstr)
    si = h_elem[2]
    if not si == sensorid:
        log.msg("Different sensorids in publish address {} and header {} - please check - aborting".format(si,sensorid))
        sys.exit()
    keylist = h_elem[3].strip('[').strip(']').split(',')
    elemlist = h_elem[4].strip('[').strip(']').split(',')
    unitlist = h_elem[5].strip('[').strip(']').split(',')
    multilist = list(map(float,h_elem[6].strip('[').strip(']').split(',')))
    if debug:
        log.msg("Packing code: {}".format(packstr))
        log.msg("keylist: {}".format(keylist))
    head_dict['SensorID'] = sensorid
    sensl = sensorid.split('_')
    head_dict['SensorName'] = sensl[0]
    head_dict['SensorSerialNum'] = sensl[1]
    head_dict['SensorRevision'] = sensl[2]
    head_dict['SensorKeys'] = ','.join(keylist)
    head_dict['SensorElements'] = ','.join(elemlist)
    head_dict['StationID'] = stid.upper()
    # possible additional data in header (because in sensor.cfg)
    #head_dict['DataPier'] = ...
    #head_dict['SensorModule'] = ...
    #head_dict['SensorGroup'] = ...
    #head_dict['SensorDescription'] = ...
    l1 = []
    l2 = []
    for idx,key in enumerate(KEYLIST):
        if key in keylist:
            l1.append(elemlist[keylist.index(key)])
            l2.append(unitlist[keylist.index(key)])
        else:
            l1.append('')
            l2.append('')
    head_dict['ColumnContents'] = ','.join(l1[1:])
    head_dict['ColumnUnits'] = ','.join(l2[1:])
    return head_dict

def interprete_data(payload, sensorid):
    """
    source:mqtt:
    """
    # future: check for json payload first

    lines = payload.split(';') # for multiple lines send within one payload
    # allow for strings in payload !!
    array = [[] for elem in KEYLIST]
    keylist = po.identifier[sensorid+':keylist']
    multilist = po.identifier[sensorid+':multilist']
    for line in lines:
        data = line.split(',')
        timear = list(map(int,data[:7]))
        #log.msg(timear)
        time = datetime(timear[0],timear[1],timear[2],timear[3],timear[4],timear[5],timear[6])
        array[0].append(time)
        for idx, elem in enumerate(keylist):
            try:
                index = KEYLIST.index(elem)
                if not elem.endswith('time'):
                    if elem in NUMKEYLIST:
                        array[index].append(float(data[idx+7])/float(multilist[idx]))
                    else:
                        array[index].append(data[idx+7])
            except:
                # might fail with list index out of range - take all others
                pass

    return np.asarray([np.asarray(elem) for elem in array],dtype=object)

def merge_two_dicts(x, y):
        z = x.copy()   # start with x's keys and values
        z.update(y)    # modifies z with y's keys and values & returns None
        return z

def on_connect(client, userdata, flags, rc, properties=None):
    global concount
    global debug
    if debug or not concount:
        log.msg("Connected with result code {}".format(str(rc)))
    global qos
    if debug or not concount:
        log.msg("Setting QOS (Quality of Service): {}".format(qos))
    if str(rc) == '0':
        pass
    elif str(rc) == '5':
        log.msg("Broker eventually requires authentication - use options -u and -P")
    # important obtain subscription from some config file or provide it directly (e.g. collector -a localhost -p 1883 -t mqtt -s wic)
    if stationid in ['all','All','ALL']:
        substring = '#'
    else:
        substring = stationid+'/#'
    if debug or not concount:
        log.msg("Subscribing to {} with qos {}".format(substring,qos))
    concount += 1
    client.subscribe(substring,qos=qos)

def on_message(client, userdata, msg):
    if not stationid in ['all','All','ALL']:
        if not msg.topic.startswith(stationid):
            return

    if not pyversion.startswith('2'):
       msg.payload= msg.payload.decode('ascii')

    global qos
    global verifiedlocation
    global stream
    global debug
    global stid
    global senslst
    global diffstruct
    digit = 0
    arrayinterpreted = False
    wsarrayinterpreted = False
    diarrayinterpreted = False
    siarrayinterpreted = False
    soarrayinterpreted = False
    dbarrayinterpreted = False
    if stationid in ['all','All','ALL']:
        stid = msg.topic.split('/')[0]
    else:
        stid = stationid
    try:
        sensorind = msg.topic.split('/')[1]
        sensorid = sensorind.replace('meta','').replace('data','').replace('dict','')
    except:
        # Above will fail if msg.topic does not contain /
        # TODO (previous version was without 1, first occurrence -> the following line should work as well although the code above is more general)
        sensorid = msg.topic.replace(stid,"",1).replace('/','').replace('meta','').replace('data','').replace('dict','')
    # define a new data stream for each non-existing sensor
    if not instrument == '':
        if not sensorid.find(instrument) > -1:
            return

    if sensorid in blacklist:
        if debug:
            print ("Sensor {} in blacklist - not collecting".format(sensorid))
        return

    ## ################################################################################
    ## ####            Eventually check for additional format libraries       #########
    ## ################################################################################
    identdic = {}

    if addlib and len(addlib) > 0:
            # Currently only one additional library is supported
            lib = addlib[0]
            #for lib in addlib:
            elemlist = []
            for elem in topic_identifiers[lib]:
                strelem = "msg.topic.{}('{}')".format(elem,topic_identifiers[lib][elem])
                elemlist.append(strelem)
            if len(elemlist) > 1:
                teststring = " and ".join(elemlist)
            else:
                teststring = "".join(elemlist)
            if eval(teststring):
                classref = class_reference.get(lib)
                #print ("1", msg.payload)
                try:
                    msg.payload, sensorid, headerline, headerdictionary, identdic = classref.GetPayload(msg.payload,msg.topic)
                except:
                    print ("Interpretation error for {}".format(msg.topic))
                    return
                #print (payload, sensorid, headerline)
                headdict[sensorid] = headerline
                headstream[sensorid] = create_head_dict(headerline,sensorid)
                headstream[sensorid] = merge_two_dicts(headstream[sensorid], headerdictionary)
                msg.topic = msg.topic+'/data'
                for el in identdic:
                    po.identifier[el] = identdic[el]

    metacheck = po.identifier.get(sensorid+':packingcode','')


    ## ################################################################################

    if msg.topic.endswith('meta') and metacheck == '':
        log.msg("Found basic header:{}".format(str(msg.payload)))
        log.msg("Quality of Service (QOS):{}".format(str(msg.qos)))
        analyse_meta(str(msg.payload),sensorid,debug=debug)
        if not sensorid in headdict:
            headdict[sensorid] = msg.payload
            # create stream.header dictionary and it here
            headstream[sensorid] = create_head_dict(str(msg.payload),sensorid)
            if debug:
                log.msg("New headdict: {}".format(headdict))
    elif msg.topic.endswith('dict') and sensorid in headdict:
        #log.msg("Found Dictionary:{}".format(str(msg.payload)))
        head_dict = headstream[sensorid]
        for elem in str(msg.payload).split(','):
            keyvaluespair = elem.split(':')
            try:
                if not keyvaluespair[1] in ['-','-\n','-\r\n']:
                    head_dict[keyvaluespair[0]] = keyvaluespair[1].strip()
            except:
                pass
        if debug:
            log.msg("Dictionary now looks like {}".format(headstream[sensorid]))
    elif msg.topic.endswith('data'):  # or readable json
        #if readable json -> create stream.ndarray and set arrayinterpreted :
        #    log.msg("Found data:", str(msg.payload), metacheck)
        if not metacheck == '':
            if 'file' in destination:
                # Import module for writing data from acquistion
                # -------------------
                #if debug:
                #    log.msg(sensorid, metacheck, msg.payload)  # payload can be split
                # Check whether header is already identified
                # -------------------
                if sensorid in headdict:
                    header = headdict.get(sensorid)
                    if sys.version_info >= (3,0):
                        metacheck = metacheck.decode()
                    if metacheck.endswith('B'):
                        packcode = metacheck.strip('<')[:-1] # drop leading < and final B
                    else:
                        packcode = metacheck.strip('<') # drop leading <
                    # temporary code - too be deleted when lemi protocol has been updated
                    if packcode.find('4cb6B8hb30f3Bc') >= 0:
                        header = header.replace('<4cb6B8hb30f3BcBcc5hL 169\n','6hLffflll {}'.format(struct.calcsize('<6hLffflll')))
                        packcode = '6hLffflll' 
                    arrayelem = msg.payload.split(';')
                    for ar in arrayelem:
                        datearray = ar.split(',')
                        # identify string values in packcode
                        # -------------------
                        # convert packcode numbers
                        cpack = []
                        for c in packcode:
                            if c.isdigit():
                                digit = int(c)
                            else:
                                cpack.extend([c] * digit)
                                digit=1
                        cpackcode = "".join(cpack)
                        for i in range(len(cpackcode)):
                            if cpackcode[-i] == 's':
                                datearray[-i] = datearray[-i]
                            elif cpackcode[-i] == 'f':
                                datearray[-i] = float(datearray[-i])
                            else:
                                datearray[-i] = int(float(datearray[-i]))
                        # pack data using little endian byte order
                        data_bin = struct.pack('<'+packcode,*datearray)
                        # Check whether destination path has been verified already
                        # -------------------
                        if not verifiedlocation:
                            if not location in [None,''] and os.path.exists(location):
                                verifiedlocation = True
                            else:
                                log.msg("File: destination location {} is not accessible".format(location))
                                log.msg("      -> please use option l (e.g. -l '/my/path') to define") 
                        if verifiedlocation:
                            filename = "{}-{:02d}-{:02d}".format(datearray[0],datearray[1],datearray[2])
                            mm.data_to_file(location, sensorid, filename, data_bin, header)
            if 'websocket' in destination:
                if not wsarrayinterpreted:
                    stream.ndarray = interprete_data(msg.payload, sensorid)
                    #streamdict[sensorid] = stream.ndarray  # to store data from different sensors
                    wsarrayinterpreted = True
                for idx,dt in enumerate(stream.ndarray[0]):
                    msecSince1970 = int((dt - datetime(1970,1,1)).total_seconds()*1000)
                    datastring = ','.join([str(val[idx]) for i,val in enumerate(stream.ndarray) if len(val) > 0 and not i == 0])
                    if debug:
                        print ("Sending {}: {},{} to webserver".format(sensorid, msecSince1970,datastring))
                    wsserver.send_message_to_all("{}: {},{}".format(sensorid,msecSince1970,datastring))
            if 'diff' in destination:
                """
                How does it work:
                - add AMOUNT of recent data to array for each fitting sensor
                - create dictionary with data streams of length "cover" - default is 5 
                - subtract all stream 
                """
                global diffcounter # used to send meta information not every step
                global prevlasttimestamp_prim
                cover = 5
                if sensorid.find(diffsens) >= 0:
                    st = diffstruct.get(sensorid, DataStream())
                    ar = interprete_data(msg.payload, sensorid)
                    ds = DataStream(header={"SensorID": sensorid}, ndarray=ar)
                    st = join_streams(ds,st)
                    st = st.cut(cover, 1, 0)
                    diffstruct[sensorid] = st
                if len(diffstruct) >= 2:
                    names = sorted(diffstruct.keys())
                    lasttimestamp_prim = diffstruct.get(names[0]).timerange()[1]
                    #print (lasttimestamp_prim, prevlasttimestamp_prim)
                    for i,name in enumerate(names):
                        if i > 0:
                            prim = diffstruct.get(names[i-1])
                            seco = diffstruct.get(name)
                            if len(prim) >= cover and len(seco) >= cover:
                                # Sub has been verified - please note: has half of the original resolution - why?
                                if not lasttimestamp_prim == prevlasttimestamp_prim:
                                    try:
                                        sub = subtract_streams(prim, seco)
                                        primname = prim.header.get("SensorID", "TEST_unkown").split("_")[1]
                                        seconame = seco.header.get("SensorID", "TEST_unkown").split("_")[1]
                                        diffname = "Diff_{}{}_0001".format(primname, seconame)
                                        keys = sub._get_key_headers(numerical=True)
                                        ilst = [KEYLIST.index(key) for key in keys]
                                        # get data line for pub
                                        time = sub.ndarray[0][-1]
                                        timestr = ",".join(str(t) for t in [time.year, time.month, time.day, time.hour, time.minute, time.second, time.microsecond])
                                        val = [sub.ndarray[i][-1] for i in ilst]
                                        if len(val) > 1:
                                            valstr = ",".join(int(val * 1000))
                                        else:
                                            valstr = int(val[0] * 1000)
                                        data = "{},{}".format(timestr, valstr)
                                        # print (data)
                                        topic = "{}/{}".format(stid, diffname)
                                        diffcounter += 1
                                        client.publish(topic + "/data", data, qos=qos)
                                        if diffcounter > 5:
                                            # now get a new sensorid name
                                            # get head line for pub
                                            keystr = "[{}]".format(",".join(keys))
                                            headtemplate = headdict.get(names[0])
                                            headtemplatelist = headtemplate.split()
                                            if keystr == headtemplatelist[3]:
                                                # then use column names and units from data header and add delta to names
                                                try:
                                                    keynamestr = "[{}]".format(",".join(["delta_{}".format(el) for el in headtemplatelist[4].replace("[","").replace("]","").split(",")]))
                                                except:
                                                    keynamestr = keystr
                                                unit = headtemplatelist[5]
                                            else:
                                                #use keys as names and arbitrary units
                                                unit = "[{}]".format(",".join(['arb'] * len(keys)))
                                                keynamestr = keystr

                                            # takeunits =  ### take from st[0]
                                            packcode = "6hL{}".format("".join(['l'] * len(keys)))
                                            multi = "[{}]".format(",".join(['1000'] * len(keys)))
                                            head = "# MagPyBin {} {} {} {} {} {} {}".format(diffname, keystr, keynamestr, unit,
                                                                                            multi, packcode,
                                                                                            struct.calcsize('<' + packcode))
                                            client.publish(topic + "/meta", head, qos=qos)
                                            diffcounter = 0
                                        prevlasttimestamp_prim = lasttimestamp_prim
                                        if debug:
                                            print(" -> diff {} published: {}".format(diffname, data))
                                    except:
                                        print (" -> diff failed")
            if 'stdout' in destination:
                if not soarrayinterpreted:
                    stream.ndarray = interprete_data(msg.payload, sensorid)
                    #streamdict[sensorid] = stream.ndarray  # to store data from different sensors
                    soarrayinterpreted = True
                for idx,dt in enumerate(stream.ndarray[0]):
                    datastring = ','.join([str(val[idx]) for i,val in enumerate(stream.ndarray) if len(val) > 0 and not i == 0])
                    log.msg("{}: {},{}".format(sensorid, dt, datastring))
            elif 'db' in destination:
                if not dbarrayinterpreted:
                    stream.ndarray = interprete_data(msg.payload, sensorid)
                    #streamdict[sensorid] = stream.ndarray  # to store data from different sensors
                    dbarrayinterpreted = True
                # create a stream.header
                if debug:
                    log.msg(stream.ndarray)
                stream.header = headstream[sensorid]
                if debug:
                    log.msg("writing header: {}".format(headstream[sensorid]))
                if revision != 'free':
                    db.write(stream,tablename="{}_{}".format(sensorid,'0001'))
                else:
                    db.write(stream)
            elif 'stringio' in destination:
                if not siarrayinterpreted:
                    stream.ndarray = interprete_data(msg.payload, sensorid)
                    #streamdict[sensorid] = stream.ndarray  # to store data from different sensors
                    siarrayinterpreted = True
                for idx,dt in enumerate(stream.ndarray[0]):
                    date = datetime.strftime(dt,"%Y-%m-%d %H:%M:%S.%f")
                    linelist = list(map(str,[dt,date]))
                    linelist.extend([str(val[idx]) for i,val in enumerate(stream.ndarray) if len(val) > 0 and not i == 0])
                    line = ','.join(linelist)
                    eol = '\r\n'
                    output.write(line+eol)
            else:
                pass
        else:
            log.msg("Non-interpreted format: {}  {}".format(msg.topic, str(msg.payload)))
    elif msg.topic and msg.topic.find('statuslog') > 0:
        # json style statusinfo is coming
        hostname = msg.topic.split('/')[-1]
        #log.msg("---------------------------------------------------------------")
        #log.msg("Receiving updated status information from {}".format(hostname))
        #log.msg("---------------------------------------------------------------")
        print ("FOUND STATUS CHANGE", telegramconf)
        statusdict = json.loads(msg.payload)
        for elem in statusdict:
            logmsg = "{}: {} - {}".format(hostname, elem, statusdict[elem])
            # For Nagios - add in marcos.log
            log.msg(logmsg)
        # For Telegram
        try:
            # try to import telegram and telegram.cfg
            ##### Add the configuration to input and marcos.cfg
            ## Please note: requires anaconda2/bin/python on my test PC
            ## !!! Requires network connection !!!
            if telegramconf and os.path.isfile(telegramconf):
                martaslog = ml(receiver='telegram')
                martaslog.receiveroptions('telegram',options={'conf':telegramconf})
                statusdict['Hostname'] = hostname
                martaslog.notify(statusdict)
        except:
            pass

        #telegram.send(msg)

    if msg.topic.endswith('meta') and 'websocket' in destination:
        # send header info for each element (# sensorid   nr   key   elem   unit)
        analyse_meta(str(msg.payload),sensorid)
        for (i,void) in enumerate(po.identifier[sensorid+':keylist']):
            jsonstr={}
            jsonstr['sensorid'] = sensorid
            jsonstr['nr'] = i
            jsonstr['key'] = po.identifier[sensorid+':keylist'][i]
            jsonstr['elem'] = po.identifier[sensorid+':elemlist'][i]
            jsonstr['unit'] = po.identifier[sensorid+':unitlist'][i]
            payload = json.dumps(jsonstr)
            wsserver.send_message_to_all('# '+payload)


def main(argv):
    broker = 'localhost'  # default
    port = 1883
    timeout=60
    altbroker = ''
    user = ''
    password = ''
    logging = 'sys.stdout'
    global destination
    destination='stdout'
    global location
    location='/tmp'
    global credentials
    credentials=''
    global offset
    offset = ''
    global dbcred
    dbcred=''
    global stationid
    stationid = 'wic'
    global stid
    stid = stationid
    global webpath
    webpath = './web'
    global webport
    webport = 8080
    global instrument
    instrument = ''
    global revision
    revision = 'fix'
    global telegramconf
    telegramconf = ''
    global addlib
    addlib = []
    global source
    source='mqtt' # projected sources: mqtt (default), wamp, mysql, postgres, etc
    global debug
    debug = False
    global output
    global headstream
    headstream = {}
    global topic_identifiers
    topic_identifiers = {}
    global class_reference
    class_reference = {}
    #global verifiedlocation
    #verifiedlocation = False
    global dictcheck
    dictcheck = False
    global socketport
    global number
    number = 1
    global qos
    qos=0
    global blacklist
    blacklist = []
    global diffsens
    global concount
    concount = 0
    conf = {}


    usagestring = 'collector.py -b <broker> -p <port> -t <timeout> -o <topic> -i <instrument> -d <destination> -v <revision> -l <location> -c <credentials> -r <dbcred> -q <qos> -u <user> -P <password> -s <source> -f <offset> -m <marcos> -n <number> -e <telegramconf> -a <addlib>'
    try:
        opts, args = getopt.getopt(argv,"hb:p:t:o:i:d:vl:c:r:q:u:P:s:f:m:n:e:a:U",["broker=","port=","timeout=","topic=","instrument=","destination=","revision=","location=","credentials=","dbcred=","qos=","debug=","user=","password=","source=","offset=","marcos=","number=","telegramconf=","addlib="])
    except getopt.GetoptError:
        print ('Check your options:')
        print (usagestring)
        sys.exit(2)
    for opt, arg in opts:
        if opt == '-h':
            print ('------------------------------------------------------')
            print ('Usage:')
            print (usagestring)
            print ('------------------------------------------------------')
            print ('Options:')
            print ('-h                             help')
            print ('-b                             set broker address: default = localhost')
            print ('-p                             set port - default is 1883')
            print ('-t                             set timeout - default is 60')
            print ('-o                             set base topic - for MARTAS this corresponds')
            print ('                               to the station ID (e.g. wic)')
            print ('                               use "-o all" to get all stationids at a specific broker')
            print ('-i                             choose instrument(s) - only sensors containing')
            print ('                               the provided string are used: ')
            print ('                               -i GSM  will access GSM90_xxx and GSM19_xyz ')
            print ('                               Default is to use all')
            print ('-d                             set destination - std.out, db, file') 
            print ('                               default is std.out') 
            print ('-v                             (no option)') 
            print ('                               if provided: meta information will be added, revision') 
            print ('                                     automatically assigned.') 
            print ('                               if not set: revision 0001 will be used, no meta.') 
            print ('-l                             set location depending on destination')
            print ('                               if d="file": provide path')
            print ('                               if d="std.out": not used')
            print ('-c                             set mqtt communication credential keyword')
            print ('-q                             set mqtt quality of service (default = 0)')            
            print ('-r                             set db credential keyword')            
            print ('-s                             source protocol of data: default is mqtt')            
            print ('                               other options:')
            print ('                               -s wamp    not yet implemented.')
            print ('-f                             offset values. Can either be "db" for')
            print ('                               applying delta values from db or a string')
            print ('                               of the following format (key:value):')
            print ('                               -f "t1:3.234,x:45674.2"')
            print ('                               other options:')
            print ('-m                             marcos configuration file ')
            print ('                               e.g. "/home/cobs/marcos.cfg"')
            print ('-n                             provide a integer number ')
            print ('                               "-d diff -i GSM": difference of two GSM will')
            print ('                                                 be calculated every n th step.')
            print ('-e                             provide a path to telegram configuration for ')
            print ('                               sending critical log changes.')
            print ('-a                             additional MQTT translation library ')
            print ('------------------------------------------------------')
            print ('Examples:')
            print ('1. Basic')
            print ('   python collector.py -b "192.168.0.100" -o wic')
            print ('2. Writing to file in directory "/my/path/"')
            print ('   python collector.py -b "192.168.0.100" -d file -l "/my/path" -o wic')
            print ('3. Writing to file and stdout')
            print ('   python collector.py -b "192.168.0.100" -d file,stdout -l "/tmp" -o wic')
            print ('4. Writing to db')
            print ('   python collector.py -b "192.168.0.100" -d db -r mydb -o wic')
            print ('   python collector.py -d "db,file" -r testdb')
            print ('5. Using configuration file')
            print ('   python collector.py -m "/path/to/marcos.cfg"')
            print ('6. Overriding individual parameters from config file')
            print ('   python collector.py -m "/path/to/marcos.cfg" -b "192.168.0.100"')
            print ('   (make sure that config is called first)')
            print ('7. Calculating differences/gradients on the fly:')
            print ('   python collector.py -d diff -i G823A -n 10')
            print ('   (will calculate the diffs of two G823A every 10th record)')
            sys.exit()
        elif opt in ("-m", "--marcos"):
            marcosfile = arg
            print ("Getting all parameters from configration file: {}".format(marcosfile))
            conf = mm.get_conf(marcosfile)
            if not conf.get('logging','') == '':
                logging = conf.get('logging').strip()
            if not conf.get('broker','') == '':
                broker = conf.get('broker').strip()
            if not conf.get('mqttport','') in ['','-']:
                port = int(conf.get('mqttport'))
            if not conf.get('mqttdelay','') in ['','-']:
                timeout = int(conf.get('mqttdelay'))
            if not conf.get('mqttuser','') in ['','-']:
                user = conf.get('mqttuser').strip()
            if not conf.get('mqttqos','') in ['','-']:
                try:
                    qos = int(conf.get('mqttqos'))
                except:
                    qos = 0
            if not conf.get('mqttcredentials','') in ['','-']:
                credentials=conf.get('mqttcredentials').strip()
            if not conf.get('blacklist','') in ['','-']:
                blacklist=conf.get('blacklist').split(',')
                blacklist = [el.strip() for el in blacklist]
            if not conf.get('station','') in ['','-']:
                stationid = conf.get('station').strip().lower()
                stid = stationid
            if not conf.get('destination','') in ['','-']:
                destination=conf.get('destination')
            if not conf.get('filepath','') in ['','-']:
                location=conf.get('filepath').strip()
            if not conf.get('dbcredentials','') in ['','-']:
                dbcred=conf.get('dbcredentials').strip()
            if not conf.get('revision','') in ['','-']:
                destination=conf.get('revision').strip()
            if not conf.get('offset','') in ['','-']:
                offset = conf.get('offset').strip()
            if not conf.get('debug','') in ['','-']:
                debug = conf.get('debug').strip()
                if debug in ['True','true']:
                    debug = True
                else:
                    debug = False
            if not conf.get('socketport','') in ['','-']:
                try:
                    socketport = int(conf.get('socketport'))
                except:
                    print('socketport could not be extracted from  marcos config file')
                    socketport = 5000
            if not conf.get('webport','') in ['','-']:
                try:
                    webport = int(conf.get('webport'))
                except:
                    print('webport could not be extracted from marcos config file')
                    webport = 8080
            if not conf.get('webpath','') in ['','-']:
                webpath = conf.get('webpath').strip()
            if not conf.get('telegramconf','') in ['','-']:
                telegramconf = conf.get('telegramconf').strip()
            if not conf.get('addlib','') in ['','-']:
                addlib = conf.get('addlib').strip().split(',')
            source='mqtt'
        elif opt in ("-b", "--broker"):
            broker = arg
        elif opt in ("-p", "--port"):
            try:
                port = int(arg)
            except:
                port = arg
        elif opt in ("-t", "--timeout"):
            try:
                timeout = int(arg)
            except:
                timeout = arg
        elif opt in ("-o", "--topic"):
            stationid = arg
        elif opt in ("-i", "--instrument"):
            instrument = arg
        elif opt in ("-s", "--source"):
            source = arg
        elif opt in ("-d", "--destination"):
            destination = arg
        elif opt in ("-l", "--location"):
            location = arg
        elif opt in ("-c", "--credentials"):
            credentials = arg
        elif opt in ("-v", "--revision"):
            revision = "free"
        elif opt in ("-r", "--dbcred"):
            dbcred = arg
        elif opt in ("-q", "--qos"):
            try:
                qos = int(arg)
            except:
                qos = 0
        elif opt in ("-u", "--user"):
            user = arg
        elif opt in ("-P", "--password"):
            password = arg
        elif opt in ("-f", "--offset"):
            offset = arg
        elif opt in ("-n", "--number"):
            number = arg
        elif opt in ("-e", "--telegramconf"):
            telegramconf = arg
        elif opt in ("-a", "--addlib"):
            addlib = arg.split(',')
        elif opt in ("-U", "--debug"):
            debug = True

    diffsens = conf.get('differencesensors',"G823")
    if debug:
        print ("collector starting with the following parameters:")
        print ("Logs: {}; Broker: {}; Topic/StationID: {}; QOS: {}; MQTTport: {}; MQTTuser: {}; MQTTcredentials: {}; Data destination: {}; Filepath: {}; DB credentials: {}; Offsets: {}".format(logging, broker, stationid, qos, port, user, credentials, destination, location, dbcred, offset))

    try:
        ##  Start Twisted logging system
        ##  ----------------------------
        if logging == 'sys.stdout':
            log.startLogging(sys.stdout)
        else:
            try:
                print (" -- Logging to {}".format(logging))
                log.startLogging(open(logging,'a'))
                log.msg("----------------")
                log.msg("  -> Logging to {}".format(logging))
            except:
                log.startLogging(sys.stdout)
                log.msg("Could not open {}. Switching log to stdout.".format(logging))
    except:
        print("Logging requires twisted module")
        sys.exit()

    try:
        ##  Eventually import additional libraries
        ##  ----------------------------
        if addlib and len(addlib) > 0:
            print ("Importing additional library")
            for lib in addlib:
                exec("from libmqtt.{} import {}".format(lib,lib))
                exec("c{} = {}()".format(lib,lib))
                class_reference[lib] = eval("c{}".format(lib))
                topic_identifiers[lib] = eval("c{}.topicidentifier".format(lib))
                print ("Imported library {}: Topic identifiers are {}".format(lib, topic_identifiers[lib]))
    except:
        pass


    if debug:
        log.msg("Logs: {}; Broker: {}; Topic/StationID: {}; QOS: {}; MQTTport: {}; MQTTuser: {}; MQTTcredentials: {}; Data destination: {}; Filepath: {}; DB credentials: {}; Offsets: {}".format(logging, broker, stationid, qos, port, user, credentials, destination, location, dbcred, offset))

    log.msg("----------------")
    log.msg(" Starting collector {}".format(__version__))
    log.msg("----------------")

    if not qos in [0,1,2]:
        qos = 1

    if 'stringio' in destination:
        #TODO seems to be unfinished - started that long ago without documentation and need to find out why ;)
        output = StringIO()
    if 'file' in destination:
        if location in [None,''] and not os.path.exists(location):
            log.msg('destination "file" requires a valid path provided as location')
            log.msg(' ... aborting ...')
            sys.exit()
    if 'websocket' in destination:
        if ws_available:
            # 0.0.0.0 makes the websocket accessable from anywhere
            global wsserver
            wsserver = WebsocketServer(socketport, host='0.0.0.0')
            wsThr = threading.Thread(target=wsThread,args=(wsserver,))
            # start websocket-server in a thread as daemon, so the entire Python program exits
            wsThr.daemon = True
            log.msg('starting WEBSOCKET on port '+str(socketport))
            wsThr.start()
            # start webserver as process, also as daemon (kills process, when main program ends)
            webPr = Process(target=webProcess, args=(webpath,webport))
            webPr.daemon = True
            webPr.start()
            log.msg('starting WEBSERVER on port '+str(webport))
        else:
            print("no webserver or no websocket-server available: remove 'websocket' from destination")
            sys.exit()
    if 'db' in destination:
        if dbcred in [None,'']:
            log.msg('destination "db" requires database credentials')
            log.msg('to create them use method "addcred"')
            log.msg('to provide use option -r like -r mydb')
            sys.exit()
        else:
            try:
                global db
                if debug:
                    log.msg("Connecting database {} at host {} with user {}".format(mpcred.lc(dbcred,'db'),mpcred.lc(dbcred,'host'),mpcred.lc(dbcred,'user')))
                #db = mysql.connect(host=mpcred.lc(dbcred,'host'),user=mpcred.lc(dbcred,'user'),passwd=mpcred.lc(dbcred,'passwd'),db=mpcred.lc(dbcred,'db'))
                db = database.DataBank(mpcred.lc(dbcred,'host'),mpcred.lc(dbcred,'user'),mpcred.lc(dbcred,'passwd'),mpcred.lc(dbcred,'db'))
                #db = dbcon.db
            except:
                log.msg('database {} at host {} with user {} could not be connected'.format(mpcred.lc(dbcred,'db'),mpcred.lc(dbcred,'host'),mpcred.lc(dbcred,'user')))
                log.msg(' ... aborting ...')
                sys.exit()

    if debug:
        log.msg("Option u: debug mode switched on ...")
        log.msg("------------------------------------")
        log.msg("Destination: {} {}".format(destination, location))

    if source == 'mqtt':
        mqttversion = int(conf.get("mqttversion", 2))
        mqttcert = conf.get("mqttcert", "")
        mqttpsk = conf.get("mqttpsk", "")
        client = connectclient(broker, port, timeout, credentials, user, password, qos, mqttcert=mqttcert, mqttpsk=mqttpsk, mqttversion=mqttversion, destinationid=dbcred, debug=debug) # dbcred is used for clientid
        client.loop_forever()

    elif source == 'wamp':
        log.msg("Not yet supported! -> check autobahn import, crossbario")
    else:
        log.msg("Additional protocols can be added in future...")


if __name__ == "__main__":
   main(sys.argv[1:])

